"""
OXASL -  Classes for representing ASL data and constructing instances from command line parameters

Copyright (c) 2008-2020 Univerisity of Oxford
"""
import collections
import json
import os
import sys
import warnings

import six
import numpy as np
import scipy.linalg

from fsl.data.image import Image

from .options import OptionCategory, OptionGroup

try:
    import oxasl_bids
except ImportError:
    oxasl_bids = None

class Options(OptionCategory):
    """
    OptionGroup which contains options for describing an ASL image
    """

    def __init__(self, fname_opt="-i"):
        OptionCategory.__init__(self, "image")
        self.fname_opt = fname_opt

    def groups(self, parser):
        group = OptionGroup(parser, "ASL data")
        group.add_option("--asldata", self.fname_opt, help="ASL data file")
        group.add_option("--iaf", help="input ASl format: diff=differenced,tc=tag-control,ct=control-tag,mp=multiphase,ve=vessel-encoded,vediff=pairwise subtracted VE data, hadamard=Hadamard time encoded data")
        group.add_option("--order", help="Data order as sequence of 2 or 3 characters: t=TIs/PLDs, r=repeats, l=labelling (tag/control/phases etc). First character is fastest varying")
        group.add_option("--tis", help="TIs (s) as comma-separated list")
        group.add_option("--tes", help="TSs (s) as comma-separated list for multi-TE ASL data")
        group.add_option("--plds", help="PLDs (s) as comma-separated list - alternative to --tis")
        group.add_option("--ntis", help="Number of TIs (for use when processing does not require actual values)", type="int")
        group.add_option("--nplds", help="Equivalent to --ntis", type="int")
        group.add_option("--rpts", help="Variable repeats as comma-separated list, one per TI/PLD", default=None)
        group.add_option("--nphases", help="For --iaf=mp, number of phases (assumed to be evenly spaced)", default=None, type="int")
        group.add_option("--nenc", help="For --iaf=ve/vediff, number of encoding cycles in original acquisition", type="int", default=8)
        group.add_option("--hadamard-size", help="Size of Hadamard encoding matrix, must be power of 2 or 12", type="int", default=8)
        group.add_option("--casl", help="Acquisition was pseudo cASL (pcASL) rather than pASL", action="store_true", default=False)
        group.add_option("--tau", "--taus", "--bolus", "--lds", help="Labelling duration (s). Can be single value or comma separated list, one per TI/PLD or per sub-bolus for Hadamard data")
        group.add_option("--slicedt", help="Timing difference between slices (s) for 2D readout", type=float, default=0.0)
        group.add_option("--sliceband", help="Number of slices per pand in multi-band setup", type=int)
        group.add_option("--artsupp", help="Arterial suppression (vascular crushing) was used", action="store_true", default=False)
        group.add_option("--ibf", help="input block format - alternative to --order for compatibility. rpt=Blocks of repeats (i.e. repeats are slowest varying), tis=Blocsk of TIs/PLDs")
        group.add_option("--calib-first-vol", help="First volume of the data is an M0 image for calibration - remainder is ASL data", action="store_true", default=False)
        return [group, ]

def summary(img, log=sys.stdout):
    """
    Write a summary of an Image to a stream

    For an AslImage the ``summary`` method is used which displays all the ASL data in human
    readable form. For generic Image objects just basic information is displayed

    :param img: fsl.data.image.Image object which may or may not be an AslImage
    :param log: Stream-like object - default is ``sys.stdout``
    """
    log.write("%s:\n" % (img.name.ljust(30)))
    if hasattr(img, "summary"):
        img.summary(log=log)

def data_order(iaf, ibf, order, multite=False):
    """
    Determine the data format and ordering from ``iaf`` and ``ibf`` options

    If ``iaf`` is not specified (None or empty string), TC pairs are specified if the
    ``order`` parameter is given and contains the ``l`` character. Otherwise differenced
    data is assumed.

    If neither ``ibf`` nor ``order`` are specified, ``ibf=rpt`` is guessed.

    If ``iaf`` is not ``diff`` and ``order`` does not include the ``l` character it is assumed
    that the encoded images (e.g. TC pairs) are the fastest varying.

    If any parameters had to be guessed (rather than inferred from other information) a
    warning is output.

    :return: Tuple of: IAF (ASL format, 'tc', 'ct', 'diff', 'quant', 've', 'vediff', 'hadamard' or 'mp'), ordering (sequence of
             2 or three chars, 'l'=labelling images, 'r'=repeats, 't'=TIs/PLDs, 'e'=TEs. The characters
             are in order from fastest to slowest varying), Boolean indicating whether the block
             format needed to be guessed
    """
    if order is not None:
        order = order.lower()

    if not iaf:
        if order is not None and "l" in order:
            # Order specified labelling images so guess TC pairs as this is most common
            warnings.warn("Data format was not specified - assuming TC pairs")
            iaf = "tc"
        elif not order:
            # No order specified  - guess differenced for compatibility with Oxford ASL
            warnings.warn("Data format was not specified - assuming differenced")
            iaf = "diff"
        else:
            # Order specified and did not include labelling images so we are entitled
            # to assume differenced data without a warning
            iaf = "diff"
    elif iaf not in ("diff", "tc", "ct", 'quant', "mp", "ve", "vediff", "hadamard"):
        raise ValueError("Unrecognized data format: iaf=%s" % iaf)

    ibf_guessed = False
    if not order:
        if not ibf:
            # Guess but defer warning until we have extracted TIs as it doesn't matter
            # for single TI data
            ibf = "rpt"
            ibf_guessed = True

        order_map = {
            "rpt" : "tr",
            "tis" : "rt",
        }
        order = order_map.get(ibf.lower(), None)
        if not order:
            raise ValueError("Unrecognized data block format: ibf=%s" % ibf)

    if iaf not in ("diff", 'quant') and "l" not in order:
        order = "l" + order

    if multite and "e" not in order:
        # TE ordering is optional since most data is single TE. Insert it at the
        # start as this seems to make most sense in terms of multi-TE acquisition
        order = "e" + order

    for char in order:
        if char not in ('l', 'r', 't', 'e'):
            raise ValueError("Unrecognized character in data ordering: '%s'" % char)

    return iaf, order, ibf_guessed

class AslImage(Image):
    """
    Subclass of fsl.data.image.Image which adds ASL structure information

    An AslImage contains information about the structure of the data enabling it to perform
    operations such as reordering and label/control differencing.

    As a *minimum* you must provide a means of determining the number of TIs/PLDs in the data.
    Specifying the data format and ordering explicitly is recommended, but a default ordering will be
    used (with a warning) if you do not.

    Ordering can be defined in two ways:

    1. Setting the ``order`` parameters to a sequence of characters (case insensitive):
      - ``l`` - Labelling images (e.g. label/control pairs, sequence of multi-phases, vessel encoding cycles)
      - ``t`` - TIs/PLDs
      - ``r`` - Repeats
      - ``e`` - TEs (optional)

      The sequence is in order from fastest varying (innermost grouping) to slowest varying (outermost
      grouping). If ``l`` is not included this describes data which is already differenced.

    2. Specifying the ``ibf`` option
      - ``ibf`` - ``rpt`` Blocked by repeats, i.e. first repeat of all TIs, followed by second repeat of all TIs...
                  ``tis`` Blocked by TIs/PLDs, i.e. all repeats of first TI, followed by all repeats of second TI...
                  When using --ibf, the labelling images (e.g. label/control pairs) are always adjacent

    The data format is defined using the ``iaf`` parameter:

      - ``iaf`` - ``tc`` = tag then control, ``ct`` = control then tag, ``mp`` = multiphase,
        ``ve`` = vessel encoded, ``vediff`` = Pairwise subtracted vessel encoded, ``diff`` = already differenced,
        ``hadamard` = Hadamard encoded, ``quant`` = already quantified

    If oxasl_bids is installed, some/all of the image metadata can be extracted from a .json sidecar if one exists

    :ivar nvols:  Number of volumes in data
    :ivar iaf:    Data format - see above
    :ivar order:  Data ordering string - see above
    :ivar ntc:    Number of labelling images in data (e.g. 2 for TC pairs, 1 for differenced data,
                  for vessel encoded or multiphase data the number of encoding cycles/phases, for Hadamard data the
                  size of the Hadamard matrix, e.g. 8)
    :ivar casl:   If True, acquisition was CASL/pCASL rather than PASL
    :ivar phases: List of phases for multiphase data (``iaf='mp'``)
    :ivar ntis:   Number of TIs/PLDs
    :ivar tis:    Optional list of TIs
    :ivar plds:    Optional list of PLDs
    :ivar have_plds: True if PLDs were provided rather than TIs
    :ivar tau:    Bolus durations - one per TI/PLD. If ``have_plds`` is True, tis are derived by adding the bolus duration to the PLDs
    :ivar rpts:   Repeats, one value per TI (may be given as a constant but always stored as list)
    :ivar slicedt: Increase in TI/PLD per slice in z dimension for 2D acquisitions (default: 0)
    :ivar sliceband: Number of slices per band for multiband acquisitions (default: None)
    :ivar artsupp: If True, data was acquired with arterial suppression
    """

    DIFFERENCED = 0
    TC_PAIRS = 1
    MULTIPHASE = 2

    def __init__(self, image, name=None, **kwargs):
        if image is None:
            raise ValueError("No image data (filename, Nibabel object or Numpy array)")

        # This is sort-of a bug in nibabel or fslpy - it passes the kwargs to the
        # nibabel.load function which does not expect extra keyword arguments. So we will extract
        # those that fslpy/nibabel might expect and only pass these
        img_kwargs = ("header", "xform", "loadData", "calcRange", "indexed", "threaded", "dataSource")
        img_args = dict([(k, v) for k, v in kwargs.items() if k in img_kwargs])

        if kwargs.pop("calib_first_vol", False):
            # First volume is an M0 calibration image - stip it off and initialize the image with the
            # remaining data
            temp_img = Image(image, name="temp", **img_args)
            img_args.pop("header", None)
            self.setMeta("calib", Image(temp_img.data[..., 0], name="calib", header=temp_img.header, **img_args))
            Image.__init__(self, temp_img.data[..., 1:], name=name, header=temp_img.header, **img_args)
        else:
            self.setMeta("calib", kwargs.pop("calib", None))
            Image.__init__(self, image, name=name, **img_args)

        # If we are initializing from a filename and we have oxasl_bids we can get metadata
        # from the BIDS sidecar
        if oxasl_bids is not None and kwargs.pop("use_bids", True) and isinstance(image, str):
            if ".nii" in image:
                sidecar_fname = image[:image.index(".nii")] + ".json"
            else:
                sidecar_fname = image + ".json"
            if os.path.exists(sidecar_fname):
                with open(sidecar_fname, "r") as f:
                    metadata = json.load(f)
                    metadata["img_shape"] = self.shape
                kwargs.update(oxasl_bids.oxasl_config_from_metadata(metadata, "asl"))

        order = kwargs.pop("order", None)
        iaf = kwargs.pop("iaf", None)
        ibf = kwargs.pop("ibf", None)
        ntis = kwargs.pop("ntis", None)
        nplds = kwargs.pop("nplds", None)
        tis = kwargs.pop("tis", None)
        tes = kwargs.pop("tes", None)
        plds = kwargs.pop("plds", None)
        rpts = kwargs.pop("rpts", kwargs.pop("nrpts", None))
        phases = kwargs.pop("phases", None)
        nphases = kwargs.pop("nphases", None)
        nenc = kwargs.pop("nenc", kwargs.get("ntc", None))
        had_size = kwargs.pop("hadamard_size", kwargs.pop("nenc", kwargs.pop("ntc", None)))

        if self.ndim not in (3, 4):
            raise ValueError("3D or 4D data expected")

        # Check for multi-TE data
        #
        # Note that for the moment we assume that multi-TEs are applied at every TI/PLD.
        # Variable repeats/taus are only per-TI. Default is one TE with value 0
        #
        # Sets the attributes tes, ntes
        if tes is not None:
            if isinstance(tes, six.string_types):
                tes = [float(te) for te in tes.split(",")]
            self.setMeta("tes", tes)
            self.setMeta("ntes", len(tes))
        else:
            self.setMeta("tes", [0,])
            self.setMeta("ntes", 1)

        # Determine the data ordering
        #
        # Sets the metadata: iaf (str), order (str)
        iaf, order, ibf_guessed = data_order(iaf, ibf, order, self.ntes > 1)
        self.setMeta("order", order)
        self.setMeta("iaf", iaf.lower())

        # Determine the number of labelling images present. This may be label/control
        # pairs, a set of multiple phases, vessel encoding cycles or already differenced data
        #
        # Sets the metadata: ntc (int), phases (list or None)
        if self.iaf in ("tc", "ct"):
            ntc = 2
        elif self.iaf == "mp":
            if phases is None and nphases is None:
                raise ValueError("Multiphase data specified but number of phases not given")
            elif phases is not None:
                if nphases is not None and nphases != len(phases):
                    raise ValueError("Number of phases is not consistent with length of phases list")
            else:
                phases = [pidx * 360 / nphases for pidx in range(nphases)]

            if isinstance(phases, six.string_types):
                phases = [float(ph) for ph in phases.split(",")]
            ntc = len(phases)
        elif self.iaf in ("ve", "vediff"):
            if nenc is None:
                raise ValueError("Vessel encoded data specified but number of encoding cycles not given")
            elif self.iaf == "ve":
                ntc = nenc
            elif nenc % 2 == 0:
                ntc = int(nenc / 2)
            else:
                raise ValueError("Subtracted vessel encoded data must have had an even number of encoding cycles")
        elif self.iaf == "hadamard":
            if had_size is None:
                raise ValueError("Hadamard encoded data specified but hadamard size not given")
            elif had_size not in (2, 4, 8, 12, 16, 32, 64):
                raise ValueError("Hadamard encoded data specified but hadamard size (%s) not a supported value" % had_size)
            ntc = had_size
        else:
            ntc = 1
        self.setMeta("ntc", ntc)
        self.setMeta("phases", phases)

        # Determine the number and type of delay images present. These may be given as TIs or PLDs.
        #
        # Internally we always have TIs, and only have PLDs as well if they were provided. If PLDs
        # were provided, the TIs are derived by adding on the bolus duration
        #
        # Sets the attributes tis (list), ntis (int), have_plds (bool), plds (list)
        if tis is not None and plds is not None:
            raise ValueError("Cannot specify PLDs and TIs at the same time")

        have_plds = False
        if nplds is not None:
            ntis = nplds
            have_plds = True

        if plds is not None:
            tis = plds
            have_plds = True

        if ntis is None and tis is None and self.iaf != "quant":
            raise ValueError("Number of TIs/PLDs not specified")
        elif tis is not None:
            if isinstance(tis, six.string_types):
                tis = [float(ti) for ti in tis.split(",")]
            if ntis is not None and len(tis) != ntis:
                raise ValueError("Number of TIs/PLDs specified as: %i, but a list of %i TIs/PLDs was given" % (ntis, len(tis)))
            if len(tis) == self.nvols and self.iaf in ("tc", "ct"):
                # Special case: we have label/control pairs with a TI/PLD for each volume.
                # However label/control TIs/PLDs must be equal in ASL and we expect a single value
                # for each pair. Check this is the case and reduce the number (note that
                # this scenario is common with BIDS datasets or JSON sidecars)
                tis_even = tis[::2]
                tis_odd = tis[1::2]
                if not np.allclose(tis_even, tis_odd):
                    raise ValueError(f"Inconsistent TIs/PLDs for label/control pairs: {tis_even} vs {tis_odd}")
                tis = tis_even
            ntis = len(tis)

        if self.iaf == "quant":
            if (ntis is not None and ntis > 1) or (tis is not None and tis != [0]):
                raise ValueError("Cannot specify TIs/PLDs for quantified data")
            ntis, tis = 1, [0]

        self.setMeta("ntis", int(ntis))
        self.setMeta("have_plds", have_plds)
        if have_plds:
            self.setMeta("plds", tis)
        else:
            self.setMeta("tis", tis)

        if ibf_guessed and len(self.tis) > 1:
            warnings.warn("Data order was not specified for multi-TI data - assuming blocks of repeats")

        # Determine the number of repeats (fixed or variable)
        #
        # Sets the attribute rpts (list, one per TI/PLD)
        if self.iaf == "quant" and self.nvols > 1:
            raise ValueError("Cannot have multi-volume data if iaf=quant")

        nvols_norpts = self.ntc * self.ntis * self.ntes
        if rpts is None:
            # Calculate fixed number of repeats
            if self.nvols % nvols_norpts != 0:
                raise ValueError("Data contains %i volumes, inconsistent with %i TIs and %i labelling images at %i TEs" % (self.nvols, self.ntis, self.ntc, self.ntes))
            rpts = [int(self.nvols / nvols_norpts)] * self.ntis
        else:
            if isinstance(rpts, six.string_types):
                rpts = [int(rpt) for rpt in rpts.split(",")]
            elif isinstance(rpts, (int, np.integer)):
                rpts = [rpts,]

            if len(rpts) == 1:
                rpts *= self.ntis
            elif len(rpts) != self.ntis:
                raise ValueError("%i TIs specified, inconsistent with %i variable repeats" % (self.ntis, len(rpts)))

            if sum(rpts) * self.ntc * self.ntes != self.nvols:
                raise ValueError("Data contains %i volumes, inconsistent with %i labelling images at %i TEs and total of %i repeats" % (self.nvols, self.ntc, self.ntes, sum(rpts)))
        self.setMeta("rpts", rpts)

        # Bolus durations should be a sequence same length as TIs/PLDs
        #
        # Sets the attributes taus (list, one per TI/PLD)
        taus = kwargs.pop("bolus", kwargs.pop("taus", kwargs.pop("tau", None)))
        if taus is None:
            taus = 1.8
        if isinstance(taus, six.string_types):
            taus = [float(tau) for tau in taus.split(",")]
        elif isinstance(taus, (float, int, np.integer)):
            taus = [float(taus),] * self.ntis

        if len(taus) == 1:
            taus = taus * ntis
        if self.iaf == "hadamard":
            # With Hadamard encoding we might have one tau per PLD as usual, but we might also have hybrid
            # protocols with 1 tau per sub-bolus. We want to expand this out to 1 per sub-bolus per PLD
            if len(taus) == self.ntis and len(taus) == self.ntc-1:
                warnings.warn("Number of LDs matches both number of PLDs and sub-boli. This is ambiguous, we will assume they are matched to PLDs")
            if len(taus) == self.ntis:
                # one per PLD - duplicate for each sub-bolus
                taus = sum([[t] * (self.ntc-1) for t in taus], [])
            elif len(taus) == self.ntc - 1:
                # One per sub-bolus - duplicate for each PLD
                taus = list(taus) * self.ntis
            elif len(taus) == self.ntis * (self.ntc-1):
                # One per PLD/sub-bolus - no need to change
                pass
            else:
                raise ValueError("%i bolus durations specified, inconsistent with %i TIs/PLDs and %i Hadamard sub-boli" % (len(taus), self.ntis, self.ntc-1))
        elif len(taus) != self.ntis:
            raise ValueError("%i bolus durations specified, inconsistent with %i TIs/PLDs" % (len(taus), self.ntis))
        self.setMeta("taus", taus)

        # Labelling type. CASL/pCASL normally associated with PLDs but can pass TIs instead.
        # However we would not expect PLDs for a non-CASL aquisition so this generates a warning
        #
        # If PLDs were provided, TIs are derived by adding the bolus duration to the PLDs
        #
        # Sets the attributes casl (bool), updates tis (list)
        casl = kwargs.pop("casl", None)
        if casl is None:
            casl = self.have_plds
        if self.have_plds:
            if not casl:
                warnings.warn("PLDs specified but aquisition was not CASL/pCASL - will treat these as TIs")
        self.setMeta("casl", casl)

        # Other acquisition parameters
        self.setMeta("slicedt", kwargs.pop("slicedt", 0))
        self.setMeta("sliceband", kwargs.pop("sliceband", None))
        self.setMeta("artsupp", kwargs.pop("artsupp", False))

    def __getattr__(self, name):
        return self.getMeta(name, None)

    @property
    def nvols(self):
        if self.ndim == 4:
            return self.shape[3]
        else:
            return 1

    @property
    def sliceband(self):
        sb = self.getMeta("sliceband", None)
        if sb is not None:
            return int(sb)
        else:
            return None

    @property
    def tis(self):
        if self.iaf == "hadmard":
            raise RuntimeError("TIs not defined for Hadamard encoded data")
        elif self.casl and self.have_plds:
            return [pld + tau for pld, tau in zip(self.plds, self.taus)]
        else:
            return self.getMeta("tis", self.getMeta("plds"))

    @property
    def nplds(self):
        return self.ntis

    @property
    def nphases(self):
        return self.getMeta("ntc", None)

    @property
    def nenc(self):
        return self.getMeta("ntc", None)

    @property
    def hadamard_size(self):
        return self.getMeta("ntc", None)

    def is_var_repeats(self):
        """
        :return: True if this data set has repeats which vary by time point
        """
        return min(self.rpts) != max(self.rpts)

    def get_vol_index(self, label_idx, ti_idx, rpt_idx, te_idx=0, order=None):
        """
        Get the volume index for a specified label, TI and repeat index

        :param label_idx: Label index starting from 0, e.g. for ``iaf=ct`` 0 would be the control image,
                          for ``iaf=mp`` 3 would be the 4th phase encoded image
        :param ti_idx: TI/PLD index, starting from 0
        :param rpt_idx: Repeat index, starting from 0
        :param te_idx: TE index for multi-TE data, starting from 0
        :param order: If specified use custom data ordering string (does not change ordering
                      within this AslImage - use ``reorder`` for that)
        """
        if order is None:
            order = self.order
        else:
            order = order.lower()
        
        if "l" not in order:
            # Need labelling image ordering even if differenced - harmless if ntc == 1
            order = "l" + order
        
        if "e" not in order:
            # Need ordering for TEs even if data is single-TE - this is harmless if ntes == 1
            order = "e" + order

        if ti_idx >= self.ntis:
            raise ValueError("Requested TI index %i but only %i TIs present" % (ti_idx, self.ntis))
        if te_idx >= self.ntes:
            raise ValueError("Requested TE index %i but only %i TEs present" % (te_idx, self.ntes))

        # Not yet found a simple way to do this which works for variable repeats
        # without crude iteration!
        ti_pos, rpt_pos, label_pos, te_pos = order.index("t"), order.index("r"), order.index("l"), order.index("e")
        its = [0, 0, 0, 0]
        vol_idx = 0
        while 1:
            ti, rpt, label, te = its[ti_pos], its[rpt_pos], its[label_pos], its[te_pos]
            if (ti, rpt, label, te) == (ti_idx, rpt_idx, label_idx, te_idx):
                return vol_idx
            its[0] += 1
            if its[0] == self._get_ncomp(order[0], ti):
                its[0] = 0
                its[1] += 1
            if its[1] == self._get_ncomp(order[1], ti):
                its[1] = 0
                its[2] += 1
            if its[2] == self._get_ncomp(order[2], ti):
                its[2] = 0
                its[3] += 1
            if its[3] == self._get_ncomp(order[3], ti):
                # Normally this is the end but we might have run out of
                # repeats but not run out of TIs
                pass
            else:
                vol_idx += 1
            if vol_idx > self.nvols:
                break

        raise ValueError("No volume for supplied TI, TE, label and repeat")

    def _get_ncomp(self, comp_id, ti):
        ret = {"t": self.ntis, "r" : self.rpts[ti], "l" : self.ntc, "e" : self.ntes}
        if comp_id in ret:
            return ret[comp_id]
        else:
            raise RuntimeError("Unknown ordering character: %s" % comp_id)

    def reorder(self, out_order=None, iaf=None, name=None):
        """
        Re-order ASL data

        The order is defined by a string in which
        r=repeats, l=labelling images, and t=tis/plds.
        The first character is the fastest varying

        So for a tag-control data set with 3 TIs and 2 repeats an order of "ltr" would be:
        TC (TI1), TC (TI2), TC (TI3), TC(TI1, repeat 2), TC(TI2 repeat 2), etc.
        """
        if out_order is None:
            out_order = self.order

        if iaf is None:
            iaf = self.iaf

        if self.iaf == "diff" and "l" in out_order:
            raise ValueError("Data is differenced but output order is not")
        elif "l" not in out_order and self.iaf != "diff":
            raise ValueError("Data is not differenced but output_order does not contain labelling")
        elif iaf != self.iaf and (iaf not in ("tc", "ct") or self.iaf not in ("tc", "ct")):
            raise ValueError("Can't change data format from '%s' to '%s'" % (self.iaf, iaf))
        
        if self.ntes > 1 and "e" not in out_order:
            out_order = "e" + out_order

        input_data = self.data
        if input_data.ndim == 3:
            input_data = input_data[..., np.newaxis]
        output_data = np.zeros(input_data.shape, dtype=input_data.dtype)
        tags = range(self.ntc)
        for te in range(self.ntes):
            for ti in range(self.ntis):
                for rpt in range(self.rpts[ti]):
                    for tag in tags:
                        if iaf != self.iaf:
                            # Change from TC to CT or vice versa
                            out_tag = 1-tag
                        else:
                            out_tag = tag
                        in_idx = self.get_vol_index(tag, ti, rpt, te)
                        out_idx = self.get_vol_index(out_tag, ti, rpt, te, order=out_order)
                        output_data[:, :, :, out_idx] = input_data[:, :, :, in_idx]

        if not name:
            name = self.name + "_reorder"
        return self.derived(image=output_data, name=name, iaf=iaf, order=out_order)

    def single_ti(self, ti_idx, order=None, name=None):
        """
        Extract the subset of data for a single TI/PLD

        FIXME will not correctly set have_plds flag in output if input has PLDs

        :param ti_idx: Index of the TI/PLD required starting from 0
        :param order: If specified the ordering string for the returned data
        :param name: Optional name for returned image. Defaults to original name with
                     suffix ``_ti<n>`` where ``<n>`` is the TI/PLD index
        :return: AslImage instance containing data only for the selected TI/PLD
        """
        if order is None:
            if self.iaf == "diff":
                order = "r"
            else:
                order = "lr"
        elif "t" in order:
            order = order.remove("t")
        order = order + "t"

        # Re-order so that TIs are together
        reordered = self.reorder(order)

        # Find the start index for this TI and the number of times it was repeated
        start = 0
        for idx in range(ti_idx):
            start += self.rpts[idx]*self.ntc*self.ntes
        nrpts = self.rpts[ti_idx]
        nvols = nrpts*self.ntc*self.ntes
        output_data = reordered.data[:, :, :, start:start+nvols]
        tis, plds = None, None
        if self.have_plds and self.plds is not None:
            plds = [self.plds[ti_idx],]
        elif not self.have_plds and self.tis is not None:
            tis = [self.tis[ti_idx],]

        if self.taus is not None:
            if self.iaf == "hadamard":
                # With Hadamard we have 1 tau per sub-bolus
                taus = self.taus[ti_idx:ti_idx+self.ntc-1]
            else:
                taus = self.taus[ti_idx]
        else:
            taus = None

        if not name:
            name = self.name + "_ti%i" % ti_idx
        return self.derived(image=output_data, name=name, order=order, tis=tis, plds=plds, taus=taus, ntis=1, rpts=nrpts)

    def diff(self, name=None):
        """
        Perform tag-control subtraction.

        Data will be reordered so the tag/control pairs are together

        Note that currently differencing is not supported for multiphase or vessel encoded data.
        For Hadamard data, the output will be the decoded sub-boli images.

        :param name: Optional name for returned image. Defaults to original name with suffix ``_diff``
        :return: AslImage instance containing differenced data
        """
        extra_kwargs = {}
        if self.iaf in ("diff", "quant"):
            # Already differenced
            return self
        elif self.iaf in ("tc", "ct"):
            output_data = np.zeros(list(self.shape[:3]) + [int(self.nvols/self.ntc)])

            # Re-order so that TC pairs are together with the tag first
            out_order = self.order.replace("l", "")
            differenced_order = out_order
            reordered = self.reorder("l" + out_order, iaf="tc").data

            for vol in range(int(self.nvols / 2)):
                tag = 2*vol
                ctrl = tag+1
                output_data[..., vol] = reordered[..., ctrl] - reordered[..., tag]
        elif self.iaf == "hadamard":
            # Hadamard decoding
            had_matrix = scipy.linalg.hadamard(self.ntc)
            output_data = np.zeros(list(self.shape[:3]) + [(self.ntc-1)*int(self.nvols/self.ntc)])

            # Re-order so that Hadamard encoding images are together and PLDs are next varying.
            # We will re-order again at the end to make sure final order is consistent with
            # the input order.
            out_order = self.order.replace("l", "")
            differenced_order = "tr"
            reordered = self.reorder("ltr").data

            # Decode each block of N Hadamard images into N-1 perfusion weighted images
            # Assumes that the Hadamard images are in adjacent volumes but we don't need to
            # worry at this point about what PLD/repeat they correspond to (see below)
            # Each column of the Hadamard matrix (excluding the first) contains +1/-1 values
            # which form a linear combination of the images in the sequence to extract the
            # decoded sub-bolus for that column number. Note that the convention is that +1
            # represents a label condition and thus should be subtracted
            for vol_idx in range(0, self.nvols):
                rpt_idx = int(vol_idx / self.ntc)
                had_idx = vol_idx % self.ntc
                for sub_bolus in range(self.ntc-1):
                    output_data[..., rpt_idx*(self.ntc-1)+sub_bolus] -= had_matrix[had_idx, sub_bolus+1]*reordered[..., vol_idx]
            output_data /= (self.ntc/2)

            # Now calculate the PLDs for the data decoded about which vary by sub-bolus
            # because for each sub-bolus, all of the sub-boli *after* it will contribute
            # to the PLD. Note that we have by now expanded the taus list so it is one
            # value per sub-bolus if it wasn't originally (not per PLD)
            #
            # For this to work with the decoded data we must have had all the original PLDs
            # together, hence why we reordered to 'ltr' ordering
            decoded_plds, decoded_taus, decoded_rpts = [], [], []
            for pld_idx in range(0, self.nplds):
                encoding_taus = self.taus[pld_idx * (self.ntc-1): (pld_idx + 1) * (self.ntc-1)]
                for subbolus_idx in range(self.ntc-1):
                    total_pld = self.plds[pld_idx]
                    for subsequent_subbolus_idx in range(subbolus_idx+1, self.ntc-1):
                        total_pld += encoding_taus[subsequent_subbolus_idx]

                    decoded_plds.append(total_pld)
                    decoded_taus.append(encoding_taus[subbolus_idx])
                    decoded_rpts.append(self.rpts[pld_idx])

            extra_kwargs["plds"] = decoded_plds
            extra_kwargs["rpts"] = decoded_rpts
            extra_kwargs["taus"] = decoded_taus
            extra_kwargs["nplds"] = self.nplds * (self.ntc-1)
        else:
            raise ValueError("Data is not tag-control pairs - cannot difference")

        if not name:
            name = self.name + "_diff"
        ret = self.derived(image=output_data, name=name, iaf="diff", order=differenced_order, **extra_kwargs)

        # It is possible that to do decoding we might have changed the data order - so put this right
        # if necessary (will do nothing if the ordering is already correct)
        return ret.reorder(out_order)

    def mean_across_repeats(self, name=None, diff=True):
        """
        Calculate the mean ASL signal across all repeats

        :param name: Optional name for returned image. Defaults to original name with suffix ``_mean``
        :param diff: If False do not difference the data before taking the mean (default: True)
        :return: Label-control subtracted AslImage with one volume per TI/PLD
        """
        if diff and self.ntc > 1:
            # Have tag-control pairs - need to subtract
            data = self.diff()
            out_order = "ret"
        elif self.ntc > 1:
            data = self
            out_order = "relt"
        else:
            data = self
            out_order = "ret"

        # Reorder so repeats are together saving original order. Note that
        # rt and tr are equivalent in the output but we want to preserve
        # whatever it was beforehand
        orig_order = data.order
        data = data.reorder(out_order)
        input_data = data.data
        if input_data.ndim == 3:
            input_data = input_data[..., np.newaxis]

        # Create output data - one repeat per ti. Note ordering of nested loops follows
        # out_order defined above
        output_data = np.zeros(list(data.shape[:3]) + [data.ntis * data.ntc * data.ntes])
        start = 0
        for ti, nrp in enumerate(data.rpts):
            for label in range(data.ntc):
                for te in range(data.ntes):
                    repeat_data = input_data[..., start:start+nrp]
                    output_data[..., te+data.ntes*(label+data.ntc*ti)] = np.mean(repeat_data, 3)
                    start += nrp

        if not name:
            name = self.name + "_mean"
        return data.derived(image=output_data, name=name, iaf=data.iaf, order=orig_order, rpts=1)

    def mean(self, name=None):
        """
        Take the mean across all volumes

        This takes a naive mean without differencing or grouping by TIs

        :param name: Optional name for returned image. Defaults to original name with suffix ``_mean``
        :return: 3D fsl.data.image.Image. Not an AslImage as timing information has been lost
        """
        meandata = self.data
        if meandata.ndim > 3:
            meandata = np.mean(meandata, axis=-1)
        if not name:
            name = self.name + "_mean"
        return Image(image=meandata, name=name, header=self.header)

    def perf_weighted(self, name=None):
        """
        Generate a perfusion weighted image by taking the mean over repeats and then
        the mean over TIs

        :param name: Optional name for returned image. Defaults to original name with suffix ``_pwi``
        :return: 3D fsl.data.image.Image. Not an AslImage as timing information lost
        """
        if not name:
            name = self.name + "_pwi"
        if self.iaf != "mp":
            mean_diffdata = self.diff().mean_across_repeats().data
            if mean_diffdata.ndim > 3:
                mean_diffdata = np.mean(mean_diffdata, axis=-1)
        else:
            # Special case for multiphase data - we cannot difference all
            # time points but we can take the difference between maximal in/out phase
            # using FFT. We do not make this part of diff() because it is not a 'proper'
            # differencing and is just a workaround for generating a PWI, e.g. for
            # registration
            meandata = self.mean_across_repeats(diff=False)
            fft = np.fft.fft(meandata.data, axis=-1)
            mean_diffdata = np.abs(fft[..., 1])

        ret = Image(image=mean_diffdata, name=name, header=self.header)
        return ret

    def split_epochs(self, epoch_size, overlap=0, time_order=None):
        """
        Split ASL data into 'epochs' of a specified size, with optional overlap
        """
        asldata = self.diff()
        if time_order is not None:
            asldata = asldata.reorder(time_order)

        epoch = 0
        epoch_start = 0
        input_data = asldata.data
        ret = []
        while 1:
            epoch_end = min(epoch_start + epoch_size, asldata.nvols)
            tis = []
            rpts = []
            for te in range(asldata.ntes):
                for ti in range(asldata.ntis):
                    for rpt in range(asldata.rpts[ti]):
                        vol_idx = asldata.get_vol_index(0, ti, rpt, te)
                        if vol_idx >= epoch_start and vol_idx < epoch_end:
                            ti_val = self.tis[ti]
                            if ti_val not in tis:
                                tis.append(ti_val)
                                rpts.append(1)
                            else:
                                idx = tis.index(ti_val)
                                rpts[idx] += 1
            epoch_data = input_data[..., epoch_start:epoch_end]
            epoch_img = AslImage(image=epoch_data,
                                 name=self.name + "_epoch%i" % epoch,
                                 iaf=asldata.iaf, order=asldata.order,
                                 tis=tis, rpts=rpts, tes=self.tes,
                                 header=self.header).mean_across_repeats()

            ret.append(epoch_img)
            epoch += 1
            epoch_start += epoch_size - overlap
            # Finish if start of next epoch is out of range or if current epoch
            # ended at the last volume
            if epoch_start >= asldata.nvols or epoch_end == asldata.nvols:
                break

        return ret

    def summary(self, log=sys.stdout):
        """
        Write a summary of the data to a file stream

        :param log: Stream-like object to write the summary to. Defaults to ``sys.stdout``
        """
        for key, value in self.metadata_summary().items():
            log.write("%s: %s\n" % (key.ljust(30), str(value)))

    def metadata_summary(self):
        """
        Generate a human-readable dictionary of metadata

        :return: Dictionary mapping human readable metadata name to value (e.g. 'Label type'
                 might map to 'Label-control pairs'). The keys and values are not guaranteed
                 to be fixed and should not be parsed - use the instance attributes instead.
        """
        md = collections.OrderedDict()
        md["Data shape"] = str(self.shape)

        if self.iaf == "tc":
            label_type = "Label-control pairs"
        elif self.iaf == "ct":
            label_type = "Control-Label pairs"
        elif self.iaf == "mp":
            label_type = "Multiphase"
        elif self.iaf == "quant":
            label_type = "Already quantified"
        elif self.iaf == "ve":
            label_type = "Vessel encoded"
        elif self.iaf == "vediff":
            label_type = "Vessel encoded (pairwise subtracted)"
        elif self.iaf == "hadamard":
            label_type = "Hadamard time-encoded"
        else:
            label_type = "Already differenced"
        md["Label type"] = label_type

        if self.iaf == "quant":
            # Other parameters basically irrelevant if data is already quantified
            return md

        if self.iaf == "mp":
            md["Phases"] = str(self.phases)
        elif self.iaf in ("ve", "vediff"):
            md["Encoding cycles"] = self.nenc

        md["Labelling"] = "CASL/pCASL" if self.casl else "PASL"
        if self.have_plds:
            md["PLDs (s)"] = str(self.plds)
        else:
            md["TIs (s)"] = str(self.tis)

        md["Repeats at each TI"] = str(self.rpts)
        if self.taus:
            md["Bolus durations (s)"] = str(self.taus)
        if self.slicedt:
            md["Time per slice (s)"] = self.slicedt
        if self.sliceband:
            md["Slices per band"] = self.sliceband
        if self.artsupp:
            md["Arterial suppression"] = "Was used"
        
        if len(self.tes) > 0:
            md["TEs (s)"] = str(self.tes)
            
        return md

    def derived(self, image, name=None, suffix="", **kwargs):
        """
        Create a derived ASL image based on this one, but with different data

        This is only possible if the number of volumes match, otherwise we cannot
        use the existing information about TIs, repeats etc. If the number of volumes
        do not match a generic Image is returned instead

        Any further keyword parameters are passed to the AslImage constructor, overriding existing
        attributes, so this can be used to create a derived image with different numbers of
        repeats, etc, provided the data is consistent with this.

        Note that the image space does not have to match, however in this case the 'header'
        kwarg should be passed to create the new image header

        :param data: Numpy data for derived image
        :param name: Name for new image (can be simple name or full filename)
        :param suffix: If name not specified, construct by adding suffix to original image name
        :return: derived AslImage. However if the AslImage constructor fails, a basic
                 fsl.data.image.Image is returned and a warning is output.
        """
        if name is None:
            name = self.name
        name = name + suffix

        derived_attrs = ["iaf", "order", "rpts", "taus", "phases", "tes",
                         "casl", "sliceband", "slicedt", "artsupp", "calib"]
        if self.have_plds:
            derived_attrs.append("plds")
            derived_attrs.append("nplds")
        else:
            derived_attrs.append("tis")
            derived_attrs.append("ntis")

        derived_kwargs = {}
        for attr in derived_attrs:
            derived_kwargs[attr] = kwargs.get(attr, getattr(self, attr, None))
        if derived_kwargs["iaf"] in ("ve", "vediff") and "nenc" not in derived_kwargs:
            derived_kwargs["nenc"] = self.nenc
        if derived_kwargs["iaf"] == "hadamard" and "hadamard_size" not in derived_kwargs:
            derived_kwargs["hadamard_size"] = self.hadamard_size

        try:
            return AslImage(image=image, name=name, header=kwargs.get("header", self.header), **derived_kwargs)
        except ValueError as exc:
            if kwargs.get("raise_on_fail", False):
                raise
            else:
                warnings.warn("AslImage.derived failed (%s) - returning fsl.data.image.Image" % str(exc))
                return Image(image=image, name=name, header=kwargs.get("header", self.header), **kwargs)



#!/bin/env python
"""
OXASL - Pre-quantification module

This is run before model fitting and should generate effectively differenced ASL data
e.g. by multiphase decoding

Copyright (c) 2008-2020 Univerisity of Oxford
"""

try:
    import oxasl_mp
except ImportError:
    oxasl_mp = None
    
def run(wsp):
    if wsp.asldata.iaf == "mp":
        if oxasl_mp is None:
            raise ValueError("Multiphase data supplied but oxasl_mp is not installed")
        wsp.sub("prequantify")
        oxasl_mp.run(wsp, output_wsp=wsp.prequantify)



"""
Plugin system for OXASL
=======================

Design
~~~~~~

A plugin is registered as a Python entry point. It must be an instance of
OxaslPlugin and provide the following information:

 - Name
 - Brief description
 - Version
 - Optional OptionCategory defining command line options supported
 - Dictionary of implementations of pipeline phases (e.g. filter, quantify...)
"""

from oxasl.options import OptionCategory

class OxaslPlugin:
    def __init__(self, name, description, version, **kwargs):
        self.name = name
        self.description = description
        self.version = version
        self.options = kwargs.pop("options", OptionCategory(self.name))
        self.run = kwargs



"""
Tests for CALIB module

FIXME need satrecov tests
FIXME need sensitivity correction tests
FIXME need test with edge correction
"""
import os

import pytest
import numpy as np

from fsl.data.image import Image

from oxasl import Workspace, reg, AslImage, brain

def get_wsp():
    wsp = Workspace(debug=True)
    wsp.sub("reg")
    wsp.sub("input")
    wsp.input.struc = Image(np.random.rand(10, 10, 10))
    wsp.input.calib = Image(np.random.rand(5, 5, 5))
    return wsp

@pytest.mark.skipif("FSLDIR" not in os.environ, reason="FSL not installed")
def test_get_ref_imgs_supplied():
    """
    Test a pre-supplied aslref is preferred to anything else
    """
    wsp = get_wsp()
    wsp.sub("input")
    user_aslref = Image(np.random.rand(5, 5, 5))
    wsp.input.aslref = user_aslref
    reg.get_ref_imgs(wsp)
    assert(np.allclose(user_aslref.data, wsp.reg.aslref.data))

@pytest.mark.skipif("FSLDIR" not in os.environ, reason="FSL not installed")
def test_get_ref_imgs_asldata_mean_tc():
    """
    Test brain extracted ASL mean is used for TC data
    """
    wsp = get_wsp()
    wsp.asldata = AslImage(np.random.rand(5, 5, 5, 4), tis=[1, 2], iaf="tc", ibf="rpt")
    reg.get_ref_imgs(wsp)
    meanasl_brain = brain.brain(wsp, wsp.asldata.mean(), thresh=0.2)
    assert(np.allclose(meanasl_brain.data, wsp.reg.aslref.data))

@pytest.mark.skipif("FSLDIR" not in os.environ, reason="FSL not installed")
def test_get_ref_imgs_asldata_mean_ct():
    """
    Test brain extracted ASL mean is used for CT data
    """
    wsp = get_wsp()
    wsp.asldata = AslImage(np.random.rand(5, 5, 5, 4), tis=[1, 2], iaf="ct", ibf="rpt")
    reg.get_ref_imgs(wsp)
    meanasl_brain = brain.brain(wsp, wsp.asldata.mean(), thresh=0.2)
    assert(np.allclose(meanasl_brain.data, wsp.reg.aslref.data))

@pytest.mark.skipif("FSLDIR" not in os.environ, reason="FSL not installed")
def test_get_ref_imgs_calib():
    """
    Test brain extracted calibration image is used for differenced data
    """
    wsp = get_wsp()
    wsp.asldata = AslImage(np.random.rand(5, 5, 5, 4), tis=[1, 2], iaf="diff", ibf="rpt")
    reg.get_ref_imgs(wsp)
    calib_brain = brain.brain(wsp, wsp.calib, thresh=0.2)
    assert(np.allclose(calib_brain.data, wsp.reg.aslref.data))



"""
FSL-compatible wrappers for Fabber tools
"""
from __future__ import absolute_import

import sys
import os

import six
import numpy as np
import nibabel as nib

from fsl.data.image import Image
from fsl.wrappers import LOAD, wrapperutils  as wutils
import fsl.utils.assertions as asrt
from fabber import Fabber, FabberException, percent_progress

from oxasl.utils import Tee

def _matching_image(base_img, img):
    if isinstance(base_img, nib.Nifti1Image):
        return img.nibImage
    elif isinstance(base_img, np.ndarray):
        return img.nibImage.get_fdata()
    else:
        return img

class _Results(dict):
    """
    Nicked from fsl.wrapperutils
    """
    def __init__(self, output):
        dict.__init__(self)
        self.__output = output

    @property
    def output(self):
        """Access the return value of the decorated function. """
        return self.__output

def fabber(options, output=LOAD, ref_nii=None, progress_log=None, **kwargs):
    """
    Wrapper for Fabber tool

    This is not a 'conventional' FSL command line tool wrapper. Rather it is
    using the Fabber Python API which interfaces to the C++ code using either
    the pure-C api and Python ctypes or its own CLI wrapper.

    The main reason for not using a conventional wrapper is that Fabber
    can take an arbitrary number of inputs including image inputs so the
    @fileOrImage type decorators don't really work well. Also you can't
    tell what image inputs you might have until you query the individual
    model for its options. All of this complexity is therefore hidden in
    the generic Fabber python API.

    Nevertheless we aim to replicate the interface of FSL tools as much as
    possible, e.g. accepting fsl.data.image.Image instances for input
    data, and respecting the LOAD special parameter to indicate which
    data items should be returned as Image instances.

    :param options: Fabber run options
    :param output: Name of output directory to put results in. The special value
                   LOAD is supported and will cause output to be returned as
                   a dictionary instead.
    :param ref_nii: Optional reference Nibabel image to use when writing output
                    files. Not required if main data is FSL or Nibabel image.
    :param progress_log: File-like stream to logging progress percentage to
    :return: Dictionary of output data items name:image. The image matches the
             type of the main input data unless this was a file in which case
             an fsl.data.image.Image is returned.
    """
    extra_search_dirs = kwargs.pop("fabber_dirs", ())
    fab = Fabber(*extra_search_dirs)

    options = dict(options)
    main_data = options.get("data", None)
    if main_data is None:
        raise ValueError("Main data not specified")

    if output != LOAD and not os.path.exists(output):
        os.makedirs(output)

    # Get a reference Nibabel image to use when generating output
    if not ref_nii:
        if isinstance(main_data, Image):
            ref_nii = main_data.nibImage
        else:
            ref_nii = Image(main_data).nibImage

    if ref_nii:
        header = ref_nii.header
        affine = ref_nii.header.get_best_affine()
    else:
        header = None
        affine = np.identity(4)

    # Replace fsl.Image objects with the underlying nibabel object. The Fabber
    # Python API can already handle Numpy arrays, nibabel images and filenames
    for key in list(options.keys()):
        value = options[key]
        if isinstance(value, Image):
            options[key] = value.nibImage

    # Streams to capture stdout and stderr and maybe send them elsewhere too
    stdout = Tee()
    stderr = Tee()

    # Deal with standard keyword arguments
    ret_exitcode = kwargs.pop("exitcode", False)
    ret_stdout = kwargs.pop("stdout", True)
    ret_stderr = kwargs.pop("stderr", False)
    log = kwargs.pop("log", {})

    stdout.add(log.get("stdout", None))
    stderr.add(log.get("stderr", None))
    if log.get("tee", False):
        stdout.add(sys.stdout)
        stderr.add(sys.stderr)

    if kwargs.pop("submit", False):
        raise ValueError("submit not supported for Fabber")

    exception = None
    cmd_output = []
    ret = _Results(cmd_output)
    try:
        ret["paramnames"] = fab.get_model_params(options)
        if log.get("cmd", None):
            log["cmd"].write(" - Using fabber:\n  core lib=%s\n  core_exe=%s\n  model libs=%s\n  model exes=%s\n" % (fab.core_lib, fab.core_exe, fab.model_libs, fab.model_exes))
            log["cmd"].write("fabber ")
            for key, value in options.items():
                if not isinstance(value, six.string_types) and not isinstance(value, (int, float)):
                    value = str(type(value))
                log["cmd"].write("--%s=%s " % (key.replace("_", "-"), value))
            log["cmd"].write("\n")
        progress_cb = None
        if progress_log:
            progress_cb = percent_progress(progress_log)
        run = fab.run(options, progress_cb)
        ret["logfile"] = run.log

        # Write output data or save it as required
        for data_name, data in run.data.items():
            nii = nib.Nifti1Image(data, header=header, affine=affine)
            nii.update_header()
            nii.header.set_data_dtype(data.dtype)
            img = Image(nii)
            if output == LOAD:
                # Return in-memory data items as the same type as image as the main data
                ret[data_name] = _matching_image(main_data, img)
            else:
                fname = os.path.join(output, data_name)
                img.save(fname)

    except FabberException as exc:
        # Error while actually running Fabber - may raise later
        # or replace with exit code
        exception = exc
        stderr.write(str(exc) + "\n")

    if ret_stdout:
        cmd_output.append(str(stdout))
    if ret_stderr:
        cmd_output.append(str(stderr))
    if ret_exitcode:
        cmd_output.append(int(exception is not None))

    if exception is not None and not ret_exitcode:
        raise exception

    return ret

@wutils.fileOrImage('mvn', 'output', 'valim', 'varim', 'mask')
@wutils.fslwrapper
def mvntool(mvn, param, **kwargs):
    """
    Wrapper for MVNTOOL.

    :param mvn: Input MVN data
    :param param: Parameter index (or name, but only if param-list is specified)

    For other arguments, see command line tool for now
    """
    asrt.assertIsNifti(mvn)

    valmap = {
        'write' : wutils.SHOW_IF_TRUE,
        'new' : wutils.SHOW_IF_TRUE,
    }

    cmd = ['mvntool', '--input={}'.format(mvn), '--param={}'.format(param)]
    cmd += wutils.applyArgStyle('--=', valmap=valmap, **kwargs)

    return cmd



"""
OXASL - Quantification using Vaby (Python based VB)

Unlike the fabber method this does not yet support PVC

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from . import multistep_fit

def run(wsp):
    # Basic non-PVC run
    multistep_fit.run(wsp.sub("basil"), impl="vaby")
    wsp.quantify_wsps.append("basil")
    


#!/bin/env python
"""
OXASL - Brain extraction

Copyright (c) 2008-2020 Univerisity of Oxford
"""

import fsl.wrappers as fsl

def brain(wsp, img, thresh=0.5):
    """
    Extract brain from wholehead image
    """
    bet_result = fsl.bet(img, seg=True, mask=False, fracintensity=thresh, output=fsl.LOAD, log=wsp.fsllog)
    return bet_result["output"]

def mask(wsp, img, thresh):
    """
    Extract brain mask from wholehead image
    """
    bet_result = fsl.bet(img, seg=False, mask=True, fracintensity=thresh, output=fsl.LOAD, log=wsp.fsllog)
    return bet_result["output_mask"]



#!/bin/env python
"""
OXASL - Module which generates perfusion stats within various ROIs

Copyright (c) 2008-2020 Univerisity of Oxford
"""
import optparse
import os

import numpy as np
import scipy.stats
import pandas as pd
from scipy.fftpack import fft, ifft

from fsl.data import atlases
from fsl.data.image import Image

from oxasl import reg, Workspace
from oxasl.options import OptionCategory, OptionGroup

class Options(OptionCategory):
    """
    Options for region analysis
    """

    def __init__(self):
        OptionCategory.__init__(self, "region_analysis")

    def groups(self, parser):
        group = OptionGroup(parser, "Region analysis")
        group.add_option("--full-region-analysis", "--region-analysis", help="Perform full region analysis using atlas ROIs", action="store_true", default=False)
        group.add_option("--uncalibrated-stats", help="Perform region analysis on uncalibrated data as well as calibrated", action="store_true", default=False)
        group.add_option("--roi-min-nvoxels", default=10, type=int,
                          help="Minimum number of relevant voxels required to report statistics")
        group.add_option("--pure-gm-thresh", "--gm-thresh", default=0.8, type=float,
                          help="Probability threshold for 'pure' grey matter")
        group.add_option("--pure-wm-thresh", "--wm-thresh", default=0.9, type=float,
                          help="Probability threshold for 'pure' white matter")
        group.add_option("--min-gm-thresh", default=0.1, type=float,
                          help="Probability threshold for a voxel to be included in GM stats")
        group.add_option("--min-wm-thresh", default=0.1, type=float,
                          help="Probability threshold for a voxel to be included in WM stats")
        group.add_option("--att-stats", help="Generate statistics for arterial transit time", action="store_true", default=False)
        group.add_option("--fuzzy-sets", help="ROI sets (groups of ROIs with total PV=1) will be modelled as fuzzy rather than binarized", action="store_true", default=False)
        group.add_option("--psf", help="Point-spread function for ROI sets. If specified, PSF will be applied to ROI sets (in ASL space) and 'fuzzy' mean used")
        group.add_option("--add-roi",
                          help="Additional ROI as binarised mask in ASL, structural, or MNI space. The name of the ROI will be the stripped filename. May be specified multiple times")
        group.add_option("--add-atlas", "--add-std-atlas", "--add-mni-atlas",
                          help="Additional regions as labelled image. If a single region is contained, the name of the ROI will be the stripped filename, otherwise use --add-std-atlas-names. May be specified multiple times (comma separated)")
        group.add_option("--add-atlas-labels", "--add-std-atlas-labels", "--add-mni-atlas-labels",
                          help="Comma separated filenames containing names of regions in atlas given in --add-atlas")
        #group.add_option("--add-standard-atlases", action="store_true", default=False,
        #                  help="Add ROIs from Harvard-Oxford cortical/subcortical atlases")
        group.add_option("--save-std-rois", "--save-mni-rois", action="store_true", default=False,
                          help="Save ROIs in standard (MNI) space")
        group.add_option("--save-struct-rois", action="store_true", default=False,
                          help="Save ROIs in structural space")
        group.add_option("--save-asl-rois", "--save-native-rois", action="store_true", default=False,
                          help="Save ROIs in ASL space")
        group.add_option("--save-asl-masks", "--save-native-masks", action="store_true", default=False,
                          help="Save binary masks in ASL space")

        return [group, ]

def _addvar(f):
    """ Add an unused 'variance' parameter to a function which doesn't use it """
    def _f_with_var(val, var):
        return f(val)
    return _f_with_var

def mean_invvarweighted(val, var):
    """ Inverse variance weighted mean (i.e. precision weighted mean) """
    if var is None:
        return None
    prec = 1 / var
    prec[~np.isfinite(prec)] = 0
    return np.sum(val * prec) / np.sum(prec)

def i2(val, var):
    """ I^2 Measure of heterogenaity """
    if var is None:
        return None
    prec = 1 / var
    prec[~np.isfinite(prec)] = 0
    n = len(val)
    mu_bar = mean_invvarweighted(val, var)

    Q = np.sum(prec * (val - mu_bar)**2)
    if Q == 0:
        i2 = 0
    else:
        i2 = (Q-(n-1))/Q

    # Negative values map to 0 (see https://www.ncbi.nlm.nih.gov/pmc/articles/PMC192859/)
    i2 = max(i2, 0)

    # Return I2 as an integer percentage
    return int(100*i2+0.5)

STATS_FNS = {
    "Mean" : _addvar(np.mean),
    "Std" : _addvar(np.std),
    "Median" : _addvar(np.median),
    "IQR" : _addvar(scipy.stats.iqr),
    "Precision-weighted mean" : mean_invvarweighted,
    "I2" : i2,
}

def get_stats_binary(wsp, stats, img, var_img, roi, suffix="",
                     ignore_nan=True, ignore_inf=True, ignore_zerovar=True,
                     min_nvoxels=10, mask=None):
    """
    Get a set of statistics for a 3D image within a binary roi

    :param img: 3D Numpy array
    :param roi: 3D Numpy array with same dimensions as img and boolean data type
    :param ignore_nan: Voxels with care NaN in img are ignored
    :param ignore_inf: Voxels which are infinite in img are ignored
    :param ignore_zerovar: Ignore voxels with zero variance
    :param min_nvoxels: If the number of voxels in the ROI is less than this number
                       (after removing Nan and infinte values) no value will be returned
    :param mask: Additional binary mask to apply to ROI

    :return: Mapping from name of statistic to value. This may be NaN or infinite depending
             on the input arguments. If the number of eligible voxels is less than min_nvoxels,
             None is returned (not NaN or zero).
    """
    if list(img.shape) != list(roi.shape):
        raise ValueError("Image must have same dimensions as ROI")
    if var_img is not None and list(var_img.shape) != list(roi.shape):
        raise ValueError("Variance image must have same dimensions as ROI")
    if mask is not None and list(mask.shape) != list(roi.shape):
        raise ValueError("Mask must have same dimensions as ROI")

    if mask is None:
        mask = np.ones(roi.shape, dtype=int)
    if ignore_nan:
        mask = np.logical_and(mask, ~np.isnan(img))
    if ignore_inf:
        mask = np.logical_and(mask, np.isfinite(img))
    if var_img is not None and ignore_zerovar:
        mask = np.logical_and(mask, var_img > 0)

    effective_roi = np.logical_and(roi, mask)

    sample_data = img[effective_roi]
    if var_img is not None:
        sample_var = var_img[effective_roi]
        sample_var[sample_var == 0] = 1e-6
    else:
        sample_var = None
    nvoxels = len(sample_data)
    stats["Nvoxels" + suffix] = nvoxels
    for stat, fn in STATS_FNS.items():
        if nvoxels < min_nvoxels:
            stats[stat + suffix] = None
        else:
            stats[stat + suffix] = fn(sample_data, sample_var)

def standardise_fuzzy(wsp, roi_set, mode="expand"):
    """
    Ensure that fuzzy ROI set has PVs that sum to 1 in every voxel

    :param roi_set: 4D Numpy array with same dimensions as img and each volume
                    containing partial volumes for each ROI in the set
    :param mode: 'expand' adds an ROI to the set containing any non-included PV.
                 'normalise' scales PVs so they sum to 1
    """
    if mode not in ("normalise", "expand"):
        raise ValueError(f"Mode {mode} not supported - expected `normalise` or `expand`")

    # Warning if normalising and ROI only has one label
    if np.shape(roi_set)[1] == 1 and mode == "normalise":
        wsp.log.write("WARNING: Only one ROI in set with mode='normalise' - all nonzero voxels will have PV=1\n")

    # Get total PV in each voxel
    roi_sum = np.sum(roi_set, axis=1, keepdims=True)

    if mode == "normalise":
        roi_set = np.where(roi_sum!=0, roi_set/roi_sum, 0)
    elif mode == "expand":
        # TODO: insert a check to see if any diff_img < 0 and raise a warning
        roi_set = np.concatenate((roi_set, 1 - roi_sum), axis=-1)
    return roi_set

def get_stats_fuzzy(wsp, stats, img, var_img, roi_set, suffix="",
                    ignore_nan=True, ignore_inf=True, ignore_zerovar=True,
                    mask=None, pv_threshold=0.):
    """
    Get a set of statistics for a set of 'fuzzy' ROIs. Unless otherwise
    specified, parameters are as for ``get_stats_binary``

    :param roi_set: 4D Numpy array with same dimensions as img and each volume
                    containing partial volumes for each ROI in the set
    :param pv_threshold: Minimum total PV for voxel to be included

    :return: Mapping from name of statistic to sequence of values, one for each ROI in the set.
             This may be NaN or infinite depending on the input arguments. 
    """
    roi_shape = list(roi_set.shape)[:3]
    if list(img.shape) != roi_shape:
        raise ValueError("Image must have same dimensions as ROI")
    if var_img is not None and list(var_img.shape) != roi_shape:
        raise ValueError("Variance image must have same dimensions as ROI")
    if mask is not None and list(mask.shape) != roi_shape:
        raise ValueError("Mask must have same dimensions as ROI")

    if mask is None:
        mask = np.ones(roi_shape, dtype=np.int32)
    if ignore_nan:
        mask = np.logical_and(mask, ~np.isnan(img))
    if ignore_inf:
        mask = np.logical_and(mask, np.isfinite(img))
    if var_img is not None and ignore_zerovar:
        mask = np.logical_and(mask, var_img > 0)

    # Only take voxels where at least one of the ROIs has non-zero percentage
    mask = np.logical_and(mask, np.sum(roi_set, axis=3) > pv_threshold)

    # Flatten ROI PVs and data into masked 2D array
    roi_array = roi_set[mask]
    g = img[mask]

    # Standardize ROI set so total PV is 1
    roi_array = standardise_fuzzy(wsp, roi_array, mode='expand')

    # Ask Jack about this???
    #if var:
    #    roi_array = np.square(roi_array)

    HT = roi_array.T
    wsp.log.write(f" - Fuzzy ROI set: condition number for transfer matrix (unweighted) = {np.linalg.cond(HT):.2f}\n")

    # Calculate roi means by linear regression
    means_lstsq, _res, _rank, _s = np.linalg.lstsq(HT@roi_array, HT@g[..., np.newaxis],
                                                   rcond=None) # None uses future default
                                                               # and silences warning

    # Note that we do not report stats for the 'background' ROI added to ensure total PV of 1
    stats["Nvoxels" + suffix] = [np.count_nonzero(roi_array[:, idx] > pv_threshold) for idx in range(roi_set.shape[-1])]
    stats["Mean" + suffix] = np.atleast_1d(np.squeeze(means_lstsq[:-1]))

    # If variance has been supplied add a precision-weighted mean
    if var_img is not None:
        V_inv = scipy.sparse.diags(1/var_img[mask])
        HT = roi_array.T @ V_inv
        wsp.log.write(f" - Fuzzy ROI set: condition number for transfer matrix (prec-weighted) = {np.linalg.cond(HT):.2f}\n")

        # Calculate roi means by linear regression
        means_lstsq, _res, _rank, _s = np.linalg.lstsq(HT@roi_array, HT@g[..., np.newaxis],
                                                       rcond=None) # None uses future default
                                                                   # and silences warning
        stats["Precision-weighted mean" + suffix] = np.atleast_1d(np.squeeze(means_lstsq[:-1]))

def apply_psf(array, psf):
    """
    Apply PSF blurring to an array (typically a binary or PV mask)

    The PSF is assumed to act only along the Z axis
    """
    if psf is None:
        return array

    # Make sure array is 4D
    array = array.astype(np.float32)
    was_3d = False
    if array.ndim == 3:
        was_3d = True
        array = array[..., np.newaxis]

    # Detect Z dimension padding by comparing size of psf and data
    n_slices = psf.shape[0]
    padding_slices = n_slices - array.shape[2]
    if padding_slices < 0 or padding_slices % 2 != 0:
        raise ValueError("Invalid padding in psf: %i slices vs %i in data (difference must be even and > 0)" % (n_slices, array.shape[2]))
    padding_slices = int(padding_slices/2)
    array = np.pad(array, [(0, 0), (0, 0), (padding_slices, padding_slices), (0, 0)], 'edge')

    # Calculate mean along z direction for each (x, y, t) to demean volume
    zmean = np.expand_dims(np.mean(array, 2), 2)
    array = array - zmean

    # Apply blurring using multiplication in Fourier domain
    fftkern = fft(psf)[np.newaxis, np.newaxis, ..., np.newaxis]
    fftvol = fft(array, axis=2)

    # Get blurred volume in Image domain and add DC term back
    blurred_array = np.real(ifft(np.multiply(fftvol, fftkern), axis=2)) + zmean

    # Unpad and Unsqueeze extra dimension if original volume was 3D
    blurred_array = blurred_array[:, :, padding_slices:-padding_slices, :]
    if was_3d:
        blurred_array = np.squeeze(blurred_array, axis=3)
    return blurred_array

def get_stats(wsp, roi_stats, roi, data_item):
    f = data_item["f"].data
    var = data_item.get("var", None)
    if var is not None:
        var = var.data

    if "fuzzy_asl" in roi:
        get_stats_fuzzy(wsp, roi_stats, f, var, roi["fuzzy_asl"], mask=data_item["mask"])
    elif "mask_asl" in roi:
        get_stats_binary(wsp, roi_stats, f, var, roi["mask_asl"], mask=data_item["mask"], min_nvoxels=wsp.roi_min_nvoxels)
    else:
        # Should never happen
        raise RuntimeError("No ASL-space ROI to get stats: %s" % str(roi))

def add_roi(wsp, rois, name, roi, threshold=0.5):
    """
    Add an ROI

    :param rois: Current list of ROIs
    :param name: Name for ROI
    :param roi: ROI image (may be binary, probabilistic, percentage...)
    :param threshold: Threshold for generating binary mask
    """
    wsp.log.write(" - %s..." % name)
    roi_space = reg.get_img_space(wsp, roi)
    roi_asl = reg.change_space(wsp, roi, "asl")
    rois.append({
        "name" : name,
        "roi_asl" : roi_asl,
        "mask_asl" : (roi_asl.data > threshold).astype(np.int32),
        "roi_%s" % roi_space : roi,
    })
    wsp.log.write("DONE\n")

def add_rois(wsp, rois, names, roi_list, header, threshold=0.5):
    """
    Add a list of ROIs

    :param rois: Current list of ROIs
    :param names: Array of ROI names
    :param roi_set: sequence of ROI Numpy arrays (may be binary, probabilistic, percentage...)
    :param header: Nifti header for ROI list arrays
    :param threshold: Threshold for generating binary mask
    """
    roi_4d = Image(np.stack(roi_list, axis=3), header=header)
    roi_space = reg.get_img_space(wsp, roi_4d)
    roi_asl = reg.change_space(wsp, roi_4d, "asl")
    if threshold:
        mask_asl = roi_asl.data > threshold
    else:
        mask_asl = roi_asl.data

    for idx, name in enumerate(names):
        wsp.log.write(" - %s..." % name)
        rois.append({
            "name" : name,
            "roi_asl" : Image(roi_asl.data[..., idx], header=roi_asl.header),
            "mask_asl" : (mask_asl[..., idx] > threshold).astype(np.int32),
            "roi_%s" % roi_space : Image(roi_list[idx], header=header),
        })
    wsp.log.write("DONE\n")

def add_roi_set(wsp, rois, set_name, names, roi_set, threshold=None):
    """
    Add an ROI set

    :param rois: Current list of ROIs
    :param names: Array of ROI names
    :param roi_set: 4D Image where volumes define disjoint masks
    :param threshold: Optional threshold for generating binary native space mask.
                      Note this is not required since roi sets can be 'fuzzy'
    """
    wsp.log.write(" - %s..." % ",".join(names))
    roi_set_space = reg.get_img_space(wsp, roi_set)
    roi_set_asl = reg.change_space(wsp, roi_set, "asl")
    if threshold:
        mask_set_asl = roi_set_asl.data > threshold
    else:
        mask_set_asl = roi_set_asl.data
    fuzzy_set_asl = apply_psf(mask_set_asl, wsp.psf)
    rois.append({
        "roi_set_name" : set_name,
        "names" : names,
        "roi_%s" % roi_set_space : roi_set,
        "roi_asl" : roi_set_asl, 
        "mask_asl" : mask_set_asl, 
        "fuzzy_asl" : fuzzy_set_asl
    })
    wsp.log.write("DONE\n")

def add_roi_set_from_fsl_atlas(wsp, rois, atlas_name, resolution=2, threshold=0.5):
    """
    Get ROIs from an FSL atlas

    :param rois: Current list of ROIs
    :param atlas_name: Name of the FSL atlas
    :param resolution: Resolution in mm
    :param threshold: Threshold for probabilistic atlases
    """
    wsp.log.write("\nAdding ROI set from standard atlas: %s (resolution=%imm, thresholding at %.2f)\n" % (atlas_name, resolution, threshold))
    registry = atlases.registry
    registry.rescanAtlases()
    desc = registry.getAtlasDescription(atlas_name)
    atlas = registry.loadAtlas(desc.atlasID, resolution=2)
    roi_set, names = [], []
    for label in desc.labels:
        roi_region = atlas.get(label=label)
        # Convert to probability
        roi_region = Image(roi_region.data / 100.0, header=roi_region.header)
        roi_set.append(roi_region.data)
        names.append(label.name)

    if wsp.fuzzy_sets or wsp.psf is not None:
        # When treating as an ROI set, do not threshold, allow it to be fuzzy
        roi_set = Image(np.stack(roi_set, axis=3), header=roi_region.header)
        add_roi_set(wsp, rois, atlas_name, names, roi_set)
    else:
        add_rois(wsp, rois, names, roi_set, roi_region.header, threshold=threshold)

def add_rois_from_3d_label_atlas(wsp, rois, atlas_img, region_names):
    """
    Get ROIs from an atlas described by a 3D label image

    This is a 3D integer image where each unizue nonzero voxel value defines an
    ROI region

    :param rois: Current list of ROIs
    :param atlas_img: Atlas label image
    :param region_names: Atlas label image region names
    """
    wsp.log.write("\nAdding ROIs from 3D atlas label image: %s\n" % (atlas_img.name))
    if atlas_img.data.ndim != 3:
        raise RuntimeError("Label atlas is not 3D: %s" % atlas_img.name)
    labels = [idx for idx in np.unique(atlas_img.data) if idx != 0]
    if len(labels) != len(region_names):
        region_names = ["Region %i" % label for label in labels]

    for name, label in zip(region_names, labels):
        roi_data = atlas_img.data.copy()
        roi_bin = (roi_data == label).astype(np.int32)
        roi = Image(roi_bin, header=atlas_img.header)
        add_roi(wsp, rois, name, roi)

def add_roi_set_from_4d_atlas(wsp, rois, atlas_name, atlas_img, region_names, threshold=0.5):
    """
    Get ROIs from an atlas described by a 4D image
 
    Each volume defines an ROI, either as binary masks or a set of fuzzy ROIs

    :param rois: Current list of ROIs
    :param atlas_img: 4D Atlas ROI image
    :param region_names: Atlas label image region names
    :param threshold: Threshold for binarizing volumes
    """
    wsp.log.write("\nAdding ROI set from 4D atlas: %s\n" % (atlas_img.name))
    if atlas_img.data.ndim != 4:
        raise RuntimeError("Atlas is not 4D: %s" % atlas_img.name)
    labels = range(atlas_img.shape[-1])
    if len(labels) != len(region_names):
        region_names = ["Region %i" % label for label in labels]

    roi_set, names = [], []
    for name, label in zip(region_names, labels):
        roi_region = atlas_img.data[..., label]
        if not wsp.fuzzy_sets and wsp.psf is None:
            add_roi(rois, name, roi_region, threshold=threshold)
        else:
            roi_set.append(roi_region.data / 100)
            names.append(label.name)

    if wsp.fuzzy_sets or wsp.psf is not None:
        roi_set = Image(np.stack(roi_set, axis=3), header=roi_region.header)
        add_roi_set(wsp, rois, roi_set, atlas_name, names)

def get_perfusion_data(wsp):
    if wsp.perfusion_wm is not None:
        wsp.log.write(" - Found partial volume corrected results - will mask ROIs using 'base' GM/WM masks (PVE thresholds: %.2f / %.2f)\n" % (wsp.min_gm_thresh, wsp.min_wm_thresh))
        data = [
            {
                "data_name" : "perfusion",
                "suffix" : "gm",
                "f" : wsp.perfusion,
                "var" : wsp.perfusion_var,
                "mask" : np.logical_and(wsp.mask.data, wsp.structural.gm_pv_asl.data > wsp.min_gm_thresh),
            },
            {
                "data_name" : "perfusion",
                "suffix" : "wm",
                "f" : wsp.perfusion_wm,
                "var" : wsp.perfusion_wm_var,
                "mask" : np.logical_and(wsp.mask.data, wsp.structural.wm_pv_asl.data > wsp.min_wm_thresh),
            },
        ]
    else:
        wsp.log.write(" - No partial volume corrected results - will mask ROIs using 'pure' GM/WM masks (PVE thresholds: %.2f / %.2f)\n" % (wsp.pure_gm_thresh, wsp.pure_wm_thresh))
        data = [
            {
                "data_name" : "perfusion",
                "suffix" : "",
                "f" : wsp.perfusion,
                "var" :  wsp.perfusion_var,
                "mask" : wsp.mask.data,
            },
            {
                "data_name" : "perfusion",
                "suffix" : "gm",
                "f" : wsp.perfusion,
                "var" :  wsp.perfusion_var,
                "mask" : np.logical_and(wsp.mask.data, wsp.structural.gm_pv_asl.data > wsp.pure_gm_thresh),
            },
            {
                "data_name" : "perfusion",
                "suffix" : "wm",
                "f" : wsp.perfusion,
                "var" :  wsp.perfusion_var,
                "mask" : np.logical_and(wsp.mask.data, wsp.structural.wm_pv_asl.data > wsp.pure_wm_thresh),
            },
        ]
    return data

def get_arrival_data(wsp):
    if wsp.arrival is None:
        return []

    # Note that for the arrival mask we also remove voxels which will be
    # eliminated from the perfusion stats because of nan/inf values or zero
    # variances. This is a bit of a hack but should help ensure that the
    # voxel set is consistent between the two measures. Perfusion can have
    # invalid values in voxels where the arrival time is valid because of
    # voxelwise calibration, however we do not expect the reverse to occur
    if wsp.perfusion_var is None:
        f_good = np.isfinite(wsp.perfusion.data)
    else:
        f_good = np.logical_and(np.isfinite(wsp.perfusion.data), wsp.perfusion_var.data > 0)
    effective_mask = np.logical_and(wsp.mask.data, f_good)

    if wsp.perfusion_wm is not None:
        wsp.log.write(" - Found partial volume corrected results - will mask ROIs using 'base' GM/WM masks (PVE thresholds: %.2f / %.2f)\n" % (wsp.min_gm_thresh, wsp.min_wm_thresh))
        arrival_data = [
            {
                "data_name" : "arrival",
                "suffix" : "gm",
                "f" : wsp.arrival,
                "var" : wsp.arrival_var,
                "mask" : np.logical_and(effective_mask, wsp.structural.gm_pv_asl.data > wsp.min_gm_thresh),
            },
            {
                "data_name" : "arrival",
                "suffix" : "wm",
                "f" : wsp.arrival_wm,
                "var" : wsp.arrival_wm_var,
                "mask" : np.logical_and(effective_mask, wsp.structural.wm_pv_asl.data > wsp.min_wm_thresh),
            },
        ]
    else:
        wsp.log.write(" - No partial volume corrected results - will mask ROIs using 'pure' GM/WM masks (PVE thresholds: %.2f / %.2f)\n" % (wsp.pure_gm_thresh, wsp.pure_wm_thresh))
        arrival_data = [
            {
                "data_name" : "arrival",
                "suffix" : "",
                "f" : wsp.arrival,
                "var" :  wsp.arrival_var,
                "mask" : effective_mask,
            },
            {
                "data_name" : "arrival",
                "suffix" : "gm",
                "f" : wsp.arrival,
                "var" :  wsp.arrival_var,
                "mask" : np.logical_and(effective_mask, wsp.structural.gm_pv_asl.data > wsp.pure_gm_thresh),
            },
            {
                "data_name" : "arrival",
                "suffix" : "wm",
                "f" : wsp.perfusion,
                "var" :  wsp.arrival_var,
                "mask" : np.logical_and(effective_mask, wsp.structural.wm_pv_asl.data > wsp.pure_wm_thresh),
            },
        ]
    return arrival_data

def run(wsp):
    """
    Entry point for OXASL

    wsp: Output data workspace
    """
    if not wsp.structural.struc:
        return

    wsp.log.write("\nRegionwise analysis\n")

    if wsp.pvwm is not None:
        wsp.structural.wm_pv_asl = wsp.pvwm
    else:
        wsp.structural.wm_pv_asl = reg.change_space(wsp, wsp.structural.wm_pv, "asl")

    if wsp.pvgm is not None:
        wsp.structural.gm_pv_asl = wsp.pvgm
    else:
        wsp.structural.gm_pv_asl = reg.change_space(wsp, wsp.structural.gm_pv, "asl")

    if wsp.pvcsf is not None:
        wsp.structural.csf_pv_asl = wsp.pvcsf
    else:
        wsp.structural.csf_pv_asl = reg.change_space(wsp, wsp.structural.csf_pv, "asl")

    wsp.pure_gm_thresh = wsp.ifnone("pure_gm_thresh", 0.8)
    wsp.pure_wm_thresh = wsp.ifnone("pure_wm_thresh", 0.9)
    wsp.min_gm_thresh, wsp.min_wm_thresh = wsp.ifnone("min_gm_thresh", 0.1), wsp.ifnone("min_wm_thresh", 0.1)

    rois = []
    wsp.log.write("\nLoading generic ROIs\n")
    add_roi(wsp, rois, "All", wsp.native.mask)
    add_roi(wsp, rois, "%i%%+GM" % (wsp.min_gm_thresh*100), wsp.structural.gm_pv_asl, wsp.min_gm_thresh)
    add_roi(wsp, rois, "%i%%+WM" % (wsp.min_wm_thresh*100), wsp.structural.wm_pv_asl, wsp.min_wm_thresh)
    add_roi(wsp, rois, "%i%%+GM" % (wsp.pure_gm_thresh*100), wsp.structural.gm_pv_asl, wsp.pure_gm_thresh)
    add_roi(wsp, rois, "%i%%+WM" % (wsp.pure_wm_thresh*100), wsp.structural.wm_pv_asl, wsp.pure_wm_thresh)
    add_roi(wsp, rois, "Cortical %i%%+GM" % (wsp.pure_gm_thresh*100), wsp.rois.cortical_gm_asl)
    add_roi(wsp, rois, "Cerebral %i%%+WM" % (wsp.pure_wm_thresh*100), wsp.rois.cerebral_wm_asl)

    wsp.log.write("\nLoading tissue PV ROI set")
    roi_set = Image(np.stack([wsp.structural.gm_pv_asl.data, wsp.structural.wm_pv_asl.data, wsp.structural.csf_pv_asl.data], axis=-1), header=wsp.structural.csf_pv_asl.header)
    add_roi_set(wsp, rois, "tissue_pvs", ["GM PV", "WM PV", "CSF PV"], roi_set)

    if wsp.full_region_analysis:
        wsp.log.write("\nPerforming full ROI region analysis")

        # Add ROIs from command line
        user_rois = [l.strip() for l in wsp.ifnone("add_roi", "").split(",") if l.strip() != ""]
        if user_rois:
            wsp.log.write("\nLoading user-specified ROIs\n")
            for fname in user_rois:
                add_roi(wsp, rois, os.path.basename(fname).split(".")[0], Image(fname), 0.5)

        add_atlas = [l.strip() for l in wsp.ifnone("add_atlas", "").split(",") if l.strip() != ""]
        atlas_labels = [l.strip() for l in wsp.ifnone("add_atlas_labels", "").split(",") if l.strip() != ""]
        for idx, fname in enumerate(add_atlas):
            if idx < len(atlas_labels):
                with open(atlas_labels[idx]) as f:
                    names = [l.strip() for l in f.readlines()]
            else:
                names = [os.path.basename(fname).split(".")[0],]
            atlas = Image(fname)
            if atlas.data.ndim == 3:
                add_rois_from_3d_label_atlas(wsp, rois, atlas, names)
            else:
                add_roi_set_from_4d_atlas(wsp, rois, fname[:fname.index(".nii")], atlas, names)

        # Add ROIs from standard atlases
        add_roi_set_from_fsl_atlas(wsp, rois, "harvardoxford-cortical", threshold=0.5)
        add_roi_set_from_fsl_atlas(wsp, rois, "harvardoxford-subcortical", threshold=0.5)

    calib_methods = list(wsp.calibration.calib_method)
    if wsp.uncalibrated_stats:
        calib_methods.insert(0, "")
    for calib_method in calib_methods:
        wsp.log.write("\nCalibration method: %s\n" % calib_method)
        if calib_method in ("", "prequantified"):
            calib_wsp = wsp.native
        else:
            calib_wsp = getattr(wsp.native, "calib_%s" % calib_method)

        wsp.log.write("\nLoading perfusion images\n")
        stats_data = get_perfusion_data(calib_wsp)
        if wsp.arrival_stats:
            wsp.log.write(" - Also generating stats for arrival data\n")
            stats_data += get_arrival_data(calib_wsp)

        # Get stats in each ROI. Add name to stats dict to make TSV output easier
        wsp.log.write("\nGetting stats - minimum of %i voxels to report in region\n" % wsp.roi_min_nvoxels)
        for data_item in stats_data:
            stats = []
            for roi in rois:
                roi_stats = {"name" : roi.get("name", None), "names" : roi.get("names", None)}
                get_stats(wsp, roi_stats, roi, data_item)

                # The ROI might just be a single ROI or it might be a 'fuzzy set' - need
                # to handle both cases
                roi_stats = dict(roi_stats)
                if roi_stats["name"] is not None:
                    roi_stats.pop("names")
                    stats.append(roi_stats)
                else:
                    for idx in range(len(roi_stats["names"])):
                        roi_stats_copy = dict(roi_stats)
                        for k in list(roi_stats.keys()):
                            if roi_stats_copy[k] is not None:
                                roi_stats_copy[k] = roi_stats_copy[k][idx]
                        roi_stats_copy["name"] = roi_stats_copy.pop("names")
                        stats.append(roi_stats_copy)

            columns = list(stats[0].keys())
            df = pd.DataFrame(stats, columns=columns)

            # Ugly hack for naming compatibility
            data_name_csv = "_" + data_item["data_name"]
            if data_name_csv == "_perfusion":
                data_name_csv = ""
            suffix = "_" + data_item["suffix"]
            suffix = "" if suffix == "_" else suffix
            if wsp.full_region_analysis:
                # Only save CSV when doing full ROI analysis
                setattr(calib_wsp, "roi_stats%s%s" % (data_name_csv, suffix), df)

            # Summary measures output as .txt files to make oxford_asl users slightly happier
            summary_mean = stats[0]["Mean"]
            summary_name = "%s%s_mean" % (data_item["data_name"], suffix)
            calib_wsp.set_item(summary_name, summary_mean, save_name=summary_name + ".txt", save_fn=str)
            if suffix == "":
                cereb_wm_mean = stats[6]["Mean"]
                cort_gm_mean = stats[5]["Mean"]
                cereb_wm_name = "%s_cerebral_wm_mean" % data_item["data_name"]
                cort_gm_name = "%s_cortical_gm_mean" % data_item["data_name"]
                calib_wsp.set_item(cereb_wm_name, cereb_wm_mean, save_name=cereb_wm_name + ".txt", save_fn=str)
                calib_wsp.set_item(cort_gm_name, cort_gm_mean, save_name=cort_gm_name + ".txt", save_fn=str)

    # Save output masks/PVE maps
    if wsp.save_asl_rois or wsp.save_asl_masks or wsp.save_struct_rois or wsp.save_std_rois:
        wsp.sub("region_rois")
        for roi in rois:
            fname = roi.get("name", roi.get("roi_set_name", "UNKNOWN")).replace(" ", "_").replace(",", "").lower()
            if wsp.save_asl_rois and "roi_asl" in roi:
                setattr(wsp.region_rois, fname + "_asl_roi", roi["roi_asl"])
            if wsp.save_asl_masks and "mask_asl" in roi:
                setattr(wsp.region_rois, fname + "_asl_mask", roi["mask_asl"])
            if wsp.save_struct_rois and "roi_struct" in roi:
                setattr(wsp.region_rois, fname + "_struct_roi", roi["roi_struct"])
            if wsp.save_std_rois and "roi_std" in roi:
                setattr(wsp.region_rois, fname + "_std_roi", roi["roi_std"])

    wsp.log.write("\nDONE\n")

class CliOptions(OptionCategory):
    """
    Options for region analysis command line interface
    """

    def __init__(self):
        OptionCategory.__init__(self, "region_analysis_cli")

    def groups(self, parser):
        group = OptionGroup(parser, "Region analysis")
        group.add_option("--oxasl-dir", help="OXASL output dir")
        group.add_option("--perfusion", help="Perfusion map in ASL native space of the oxasl output")
        group.add_option("--output", help="Output directory", default="roicli")
        return [group]

def main():
    """
    Command line entry point for running 'compatible' region analysis
    """
    parser = optparse.OptionParser()
    parser.add_option("--oxasl-dir", help="OXASL output dir")
    parser.add_option("--perfusion", help="Perfusion map in ASL native space of the oxasl output")
    parser.add_option("--perfusion-var", help="Perfusion variance map in ASL native space of the oxasl output")
    parser.add_option("--arrival", help="ATT map in ASL native space of the oxasl output")
    parser.add_option("--arrival-var", help="ATT variance map in ASL native space of the oxasl output")
    parser.add_option("--output", "-o", help="Output directory", default="roicli")

    groups = Options().groups(parser)
    for group in groups:
        parser.add_option_group(group)

    options, _ = parser.parse_args()

    wsp = Workspace(savedir=options.output)
    wsp.sub("native")
    wsp.native.perfusion = Image(options.perfusion)
    if options.perfusion_var:
        wsp.native.perfusion_var = Image(options.perfusion_var)
    if options.arrival:
        wsp.native.arrival = Image(options.arrival)
    if options.arrival_var:
        wsp.native.arrival_var = Image(options.arrival)

    for k, v in options.__dict__.items():
        setattr(wsp, k, v)

    wsp.native.mask = Image(os.path.join(options.oxasl_dir, "output", "native", "mask"))
    wsp.full_region_analysis = True

    copy = {
        "reg" : ["aslref", "strucref", "stdref", "asl2struc.mat",
                 "struc2std", "struc2asl.mat", "std2struc"],
        "rois" : ["cortical_gm_asl", "cerebral_wm_asl"],
        "structural" : ["struc"],
        "calibration" : []
    }

    for wsp_name, fnames in copy.items():
        subwsp = wsp.sub(wsp_name)
        for fname in fnames:
            fpath = os.path.join(options.oxasl_dir, wsp_name, fname)
            if fname.endswith(".mat"):
                obj = np.loadtxt(fpath)
            else:
                obj = Image(fpath)
            attr = fname[:fname.index(".")] if "." in fname else fname
            setattr(subwsp, attr, obj)

    wsp.pvgm = Image(os.path.join(options.oxasl_dir, "structural", "gm_pv_asl"))
    wsp.pvwm = Image(os.path.join(options.oxasl_dir, "structural", "wm_pv_asl"))
    wsp.pvcsf = Image(os.path.join(options.oxasl_dir, "structural", "csf_pv_asl"))
    wsp.calibration.calib_method = [""]

    run(wsp)
    wsp.reg = None
    wsp.rois = None
    wsp.structural = None
    wsp.calibration = None
    wsp.input = None
    wsp.pvgm, wsp.pvwm, wsp.pvcsf = None, None, None




#!/bin/env python
"""
OXASL - Registration for ASL data

Michael Chappell, IBME QuBIc & FMRIB Image Analysis Groups

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from __future__ import unicode_literals

import os

import numpy as np

import fsl.wrappers as fsl

from oxasl import reg
from oxasl.reporting import LineGraph

def run(wsp):
    """
    Calculate motion correction transforms for ASL data

    Note simple motion correction of multi-volume calibration data is done in preprocessing.

    The reference volume for motion correction is the calibration image, if supplied, or
    otherwise the middle volume of the ASL data is used.

    If the calibration image is used, the inverse of the middle ASL volume -> calibration
    transform is applied to each transform matrix. This ensures that the middle volume of
    the ASL data is unchanged and interpolation on the other volumes is also minimised.
    In this case, all calibration images are also themselves transformed to bring them in
    to ASL middle volume space.

    Required workspace attributes
    -----------------------------

     - ``asldata`` : ASL data image

    Optional workspace attributes
    -----------------------------

     - ``calib``    : Calibration image

    Updated workspace attributes
    ----------------------------

     - ``asldata_mc_mats`` : Sequence of matrices giving motion correction transform for each ASL volume
     - ``asl2calib``       : ASL->calibration image transformation
     - ``calib2asl``       : Calibration->ASL image transformation
    """
    if not wsp.mc:
        return

    wsp.sub("moco")
    wsp.log.write("\nCalculating Motion Correction\n")
    if wsp.preproc.asldata.ndim < 4:
        wsp.log.write(" - Skipping, data is 3D\n")
        wsp.moco.output = wsp.preproc.asldata
        return

    nvols = wsp.preproc.asldata.shape[3]
    # If available, use the calibration image as reference since this will be most consistent if the data has a range
    # of different TIs and background suppression etc. This also removes motion effects between asldata and calibration image
    if wsp.input.aslref is not None:
        wsp.log.write(" - Using user-specified aslref as reference\n")
        ref_source = "User specified: %s" % (wsp.input.aslref.name)
        wsp.moco.ref = wsp.input.aslref
        wsp.moco.input = wsp.preproc.asldata
        mcflirt_result = fsl.mcflirt(wsp.moco.input, reffile=wsp.moco.ref, out=fsl.LOAD, mats=fsl.LOAD, log=wsp.fsllog)
        mats = [mcflirt_result[os.path.join("out.mat", "MAT_%04i" % vol)] for vol in range(nvols)]
    elif wsp.preproc.calib is not None:
        wsp.log.write(" - Using calibration image as reference\n")
        ref_source = "Calibration image"
        wsp.moco.ref = wsp.preproc.calib
        wsp.moco.input = wsp.preproc.asldata
        mcflirt_result = fsl.mcflirt(wsp.moco.input, reffile=wsp.moco.ref, out=fsl.LOAD, mats=fsl.LOAD, log=wsp.fsllog)
        mats = [mcflirt_result[os.path.join("out.mat", "MAT_%04i" % vol)] for vol in range(nvols)]

        # To reduce interpolation of the ASL data change the transformations so that we end up in the space of the central volume of asldata
        wsp.moco.asl2calib = mats[int(float(len(mats))/2)]
        wsp.moco.calib2asl = np.linalg.inv(wsp.moco.asl2calib)
        mats = [np.dot(wsp.moco.calib2asl, mat) for mat in mats]
        wsp.log.write("   ASL middle volume->Calib:\n%s\n" % str(wsp.moco.asl2calib))
        wsp.log.write("   Calib->ASL middle volume:\n%s\n" % str(wsp.moco.calib2asl))
    else:
        wsp.log.write(" - Using ASL data middle volume as reference\n")
        ref_source = "ASL data middle volume: %i" % int(float(wsp.preproc.asldata.shape[3])/2)
        wsp.moco.input = wsp.preproc.asldata
        mcflirt_result = fsl.mcflirt(wsp.moco.input, out=fsl.LOAD, mats=fsl.LOAD, log=wsp.fsllog)
        mats = [mcflirt_result[os.path.join("out.mat", "MAT_%04i" % vol)] for vol in range(nvols)]

    # Convert motion correction matrices into single (4*nvols, 4) matrix - convenient for writing
    # to file, and same form that applywarp expects. Also save output ASL data + PWI
    wsp.moco.mc_mats = np.concatenate(mats, axis=0)
    output = reg.transform(wsp, wsp.moco.input, trans=wsp.moco.mc_mats, ref=wsp.preproc.aslspace)
    wsp.moco.output = wsp.asldata.derived(output)
    try:
        wsp.moco.pwi = wsp.moco.output.perf_weighted()
    except:
        # Ignore - not all data can generate a PWI
        pass

    page = wsp.report.page("moco")
    page.heading("Motion correction", level=0)
    page.heading("Reference volume", level=1)
    page.text(ref_source)
    page.heading("Motion parameters", level=1)
    moco_params = [reg.get_transform_params(mat) for mat in mats]
    trans = [p[0] for p in moco_params]
    abstrans = np.fabs(trans)
    rot = [p[1] for p in moco_params]
    absrot = np.fabs(rot)
    page.table([
        ["Mean translation", "%.3g mm" % np.mean(trans)],
        ["Translation std.dev.", "%.3g mm" % np.std(trans)],
        ["Absolute maximum translation", "%.3g mm (volume %i)" % (np.max(abstrans), np.argmax(abstrans))],
        ["Mean rotation", "%.3g \N{DEGREE SIGN}" % np.mean(rot)],
        ["Rotation std.dev.", "%.3g \N{DEGREE SIGN}" % np.std(rot)],
        ["Absolute maximum rotation", "%.3g \N{DEGREE SIGN} (volume %i)" % (np.max(absrot), np.argmax(absrot))],
    ])
    page.image("moco_trans", LineGraph(trans, "Volume number", "Translation (mm)"))
    page.image("moco_rot", LineGraph(rot, "Volume number", "Rotation relative to reference (\N{DEGREE SIGN})"))



"""
Modelling support for the QUASAR ASL sequence
---------------------------------------------

Like pCASL, the QUASAR sequence consists of a set of label-control image
pairs acquired at single or multiple delay times, and with the potential
for repeated measurements at each PLD. However each label-control pair is
additionally acquired 7 times, each with a different combination of
arterial flow suppression ('Crusher') gradients. The total number of volumes
in the data set is therefore 2 x 7 x Number of PLDs x Number of repeats.

The standard QUASAR sequence uses 13 PLDs and 6 repeats giving a total of
1092 volumes. TC pairs alternate for the 7 cycles, and one TI is processed
(with all its repeats) at a time (--ibf=tis)

The 'F' form of QUASAR data is averaged over repeats and with ordered so that
the TC pairs for each cycle (at all PLDs) are together (effectively --ibf=cycles)

For processing we require to convert to the 'F' form which would amount to::

  asldata.mean_across_repeats().reorder("ltc")

(c is a ordering character which refers to 'cycles' - this is not yet implemented
in AslImage)

Data in the 'F' form is then differenced to yield a dataset with ordering "tc" and the
final low flip angle cycle is discarded. The resulting dataset will have 13x6=78 
volumes. A mask may be generated from the mean of the control images.

In the calibration step, Fabber is run on the 91 control images using the satrecov
model (including the low-flip angle phase) to obtain the M0 image for calibration.
An edge correction step may be performed using a median filter on the edge voxels.
(alternatively ASL_CALIB can be called instead on the control images using the
satrecov method - note that this is not yet implemented in OXASL)

If PVC required, there is now a registration and segmentation of a structural 
image (or FSL_ANAT output). OXASL should have all this done already. The endpoint
shold be GM and WM PV estimates in ASL space. Interestingly the T1t output
from the satrecov calibration is used as the ASL space reference image.

Model based analysis runs the subtracted data on the quasar model with considerable
complexity in prior specification!

Model-free analysis uses asl_mfree.
"""



"""
OXASL - ASL data preprocessing

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from __future__ import print_function

import sys

import fsl.wrappers as fsl
from fsl.data.image import Image

import scipy
import numpy as np

from oxasl import Workspace, image
from oxasl.options import AslOptionParser, OptionCategory, OptionGroup, GenericOptions

def run(wsp):
    wsp.log.write("\nPre-processing input images\n")
    wsp.sub("preproc")
    asldata_reorient = _reorient(wsp, wsp.input.asldata)
    wsp.preproc.asldata = wsp.input.asldata.derived(asldata_reorient.data)
    wsp.preproc.aslspace = wsp.preproc.asldata.mean()
    try:
        wsp.preproc.pwi = wsp.asldata.perf_weighted()
    except:
        # Ignore - not all data can generate a PWI
        pass

    if wsp.calib_first_vol and wsp.input.calib is None:
        wsp.input.calib = wsp.asldata.calib

    wsp.preproc.calib = _single_volume(wsp, wsp.input.calib)
    wsp.preproc.cref = _single_volume(wsp, wsp.input.cref)
    wsp.preproc.cact = _single_volume(wsp, wsp.input.cact)
    wsp.preproc.cblip = _single_volume(wsp, wsp.input.cblip)

def _reorient(wsp, img):
    if wsp.noreorient:
        wsp.log.write(" - NOT reorienting input data to standard orientation - need to check registration is OK\n")
        return img
    else:
        wsp.log.write(" - Reorienting to standard orientation: %s\n" % img.name)
        ret = fsl.fslreorient2std(img, output=fsl.LOAD)
        return ret["output"]

def _single_volume(wsp, img, moco=True, discard_first=True):
    """
    Convert a potentially 4D image into a single 3D volume

    :param moco: If True, perform basic motion correction
    :param discard_first: If True, discard first volume if nvols > 1

    """
    if img is not None:
        wsp.log.write(" - Pre-processing image: %s\n" % img.name)
        if img.ndim == 4:
            if discard_first and img.shape[3] > 1:
                wsp.log.write("   - Removing first volume to ensure data is in steady state\n")
                img = Image(img.data[..., 1:], header=img.header)

        if img.ndim == 4:
            if moco and img.shape[3] > 1:
                if moco:
                    wsp.log.write("   - Motion correcting\n")
                    img = fsl.mcflirt(img, out=fsl.LOAD, log=wsp.fsllog)["out"]

            wsp.log.write("   - Taking mean across volumes\n")
            img = Image(np.mean(img.data, axis=-1), header=img.header)

        return _reorient(wsp, img)
    else:
        return None

def preprocess(wsp):
    """
    Basic preprocessing of ASL data

    :param wsp: Workspace object

    Required workspace attributes
    -----------------------------

     - ``asldata`` : AslImage containing data to preprocess


    Optional workspace attributes
    -----------------------------

     - ``diff`` : If True, perform tag/control subtraction
     - ``reorder`` : If specified, reorder the data using this ordering string
     - ``mc`` : If True, perform motion correction using mcflirt
     - ``smooth`` : If True, perform smoothing
     - ``fwhm`` : If smooth=True, the full-width-half-maximum for the smoothing kernel
     - ``ref`` : If mc=True, a reference image for motion correction (default is to use the middle volume)

    Workspace attributes updated
    -----------------------------

     - ``asldata_preproc`` : preprocessed AslImage instance
    """
    wsp.log.write("ASL preprocessing...\n")

    wsp.asldata_preproc = wsp.asldata

    if wsp.diff:
        if wsp.asldata.iaf == "hadamard":
            wsp.log.write("  - Hadamard decoding\n")
        else:
            wsp.log.write("  - Tag-control subtraction\n")
        wsp.asldata_preproc = wsp.asldata_preproc.diff()

    if wsp.reorder:
        wsp.log.write("  - Re-ordering to %s\n" % wsp.reorder)
        if "l" in wsp.reorder.lower() and wsp.diff:
            wsp.reorder = wsp.reorder.replace("l", "")
        wsp.asldata_preproc = wsp.asldata_preproc.reorder(wsp.reorder)

    if wsp.mc:
        wsp.log.write("  - Motion correction\n")
        mcimg = fsl.mcflirt(wsp.asldata_preproc, cost="mutualinfo", out=fsl.LOAD, reffile=wsp.ref)["out"]
        wsp.asldata_preproc = wsp.asldata_preproc.derived(mcimg.data, suffix="_mc")

    if wsp.smooth:
        wsp.sigma = round(wsp.fwhm / 2.355, 2)
        wsp.log.write("  - Spatial smoothing with FWHM: %f (sigma=%f)\n" % (wsp.fwhm, wsp.sigma))
        smoothed = scipy.ndimage.gaussian_filter(wsp.asldata_preproc.data, sigma=wsp.sigma)
        wsp.asldata_preproc = wsp.asldata_preproc.derived(smoothed, suffix="_smooth")

    wsp.log.write("DONE\n\n")

class AslPreprocOptions(OptionCategory):
    """
    OptionCategory which contains options for preprocessing ASL data
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "preproc")

    def groups(self, parser):
        group = OptionGroup(parser, "Preprocessing")
        group.add_option("--diff", help="Perform tag-control subtraction / Hadamard decoding", action="store_true", default=False)
        group.add_option("--smooth", help="Spatially smooth data", action="store_true", default=False)
        group.add_option("--fwhm", help="FWHM for spatial filter kernel", type="float", default=6)
        group.add_option("--mc", help="Motion correct data", action="store_true", default=False)
        group.add_option("--ref", help="Optional reference image for motion correction", type="image", default=None)
        group.add_option("--reorder", help="Re-order data in specified order")
        return [group, ]

def main():
    """
    Entry point for command line tool
    """
    try:
        parser = AslOptionParser(usage="asl_preproc -i <filename> [options]")
        parser.add_category(GenericOptions(output_type="file"))
        parser.add_category(image.Options())
        parser.add_category(AslPreprocOptions())

        options, _ = parser.parse_args()
        if not options.output:
            options.output = "oxasl_preproc_out"
        wsp = Workspace(auto_asldata=True, **vars(options))

        print("")
        wsp.asldata.summary()
        preprocess(wsp)
        wsp.asldata_preproc.save(options.output)

    except ValueError as exc:
        sys.stderr.write("ERROR: " + str(exc) + "\n")
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/bin/env python
"""
OXASL - Sensitivity correction module

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from __future__ import unicode_literals

import numpy as np

from fsl.data.image import Image

from oxasl import reg
from oxasl.options import OptionCategory, OptionGroup
from oxasl.reporting import LightboxImage

class Options(OptionCategory):
    """
    Options for corrections of the input data
    """

    def __init__(self):
        OptionCategory.__init__(self, "corrections")

    def groups(self, parser):
        ret = []

        g = OptionGroup(parser, "Sensitivity correction")
        g.add_option("--cref", help="Reference image for sensitivity correction", type="image")
        g.add_option("--cact", help="Image from coil used for actual ASL acquisition (default: calibration image - only in longtr mode)", type="image")
        g.add_option("--isen", help="User-supplied sensitivity correction in ASL space")
        g.add_option("--senscorr-auto", "--senscorr", help="Apply automatic sensitivity correction using bias field from FAST", action="store_true", default=False)
        g.add_option("--senscorr-off", help="Do not apply any sensitivity correction", action="store_true", default=False)
        ret.append(g)

        return ret

def run(wsp):
    """
    Get sensitivity correction image

    Required workspace attributes
    -----------------------------

     - ``asldata`` : ASL data

    Optional workspace attributes
    -----------------------------

     - ``isen`` : User supplied sensitivity image
     - ``cact`` : Calibration image. Used in conjunction with ``cref`` to calculate sensitivity map
     - ``calib`` : Calibration image. Used as alternative to cact provided ``mode`` is ``longtr``
     - ``cref`` : Calibration reference image
     - ``senscorr_auto`` : If True, automatically calculate sensitivity correction using FAST
     - ``senscorr_off`` If True, do not apply sensitivity correction

    Updated workspace attributes
    ----------------------------

     - ``sensitivity``    : Sensitivity correction image in ASL space
    """
    wsp.log.write("\nCalculating Sensitivity correction\n")
    sensitivity = None
    bias = None
    if wsp.senscorr_off:
        wsp.log.write(" - Sensitivity correction disabled\n")
    elif wsp.isen is not None:
        wsp.log.write(" - Sensitivity image supplied by user\n")
        sensitivity = wsp.isen
    elif wsp.cact is not None and wsp.cref is not None:
        wsp.log.write(" - Sensitivity image calculated from calibration actual and reference images\n")
        cref_data = np.copy(wsp.cref.data)
        cref_data[cref_data == 0] = 1
        sensitivity = Image(wsp.cact.data.astype(np.float32) / cref_data, header=wsp.calib.header)
    elif wsp.calib is not None and wsp.cref is not None:
        if wsp.ifnone("mode", "longtr") != "longtr":
            raise ValueError("Calibration reference image specified but calibration image was not in longtr mode - need to provided additional calibration image using the ASL coil")
        wsp.log.write(" - Sensitivity image calculated from calibration and reference images\n")
        cref_data = np.copy(wsp.cref.data)
        cref_data[cref_data == 0] = 1
        sensitivity = Image(wsp.calib.data.astype(np.float32) / cref_data, header=wsp.calib.header)
    elif wsp.senscorr_auto and wsp.structural.bias is not None:
        wsp.log.write(" - Sensitivity image calculated from bias field\n")
        bias = reg.change_space(wsp, wsp.structural.bias, "asl")
        sensitivity = Image(np.reciprocal(bias.data), header=bias.header)
    else:
        wsp.log.write(" - No source of sensitivity correction was found\n")

    if sensitivity is not None:
        sdata = sensitivity.data
        sdata[sdata < 1e-12] = 1
        sdata[np.isnan(sdata)] = 1
        sdata[np.isinf(sdata)] = 1
        wsp.sub("senscorr")
        wsp.senscorr.sensitivity = Image(sdata, header=sensitivity.header)

        page = wsp.report.page("sensitivity")
        page.heading("Sensitivity correction", level=0)
        page.heading("Sensitivity map", level=1)
        page.image("sensitivity", LightboxImage(wsp.senscorr.sensitivity))

    if bias is not None:
        wsp.senscorr.bias = bias



"""
OXASL - Structural data module

Copyright (c) 2008-2020 Univerisity of Oxford
"""
import os
import glob

import numpy as np

import fsl.wrappers as fsl
from fsl.data.image import Image
from fsl.utils.path import PathError

from oxasl.options import OptionCategory, OptionGroup
from oxasl.reporting import LightboxImage

class Options(OptionCategory):
    """
    OptionGroup which contains options for describing a structural image
    """

    def __init__(self):
        OptionCategory.__init__(self, "struc")

    def groups(self, parser):
        group = OptionGroup(parser, "Structural data")
        group.add_option("--struc", "-s", help="Structural image", type="image", default=None)
        group.add_option("--struc-brain", "--sbet", "--struc-bet", "--sbrain", type="image", help="Structural image (brain extracted)", default=None)
        group.add_option("--struc2asl", help="Structural->ASL transformation matrix", type="matrix", default=None)
        group.add_option("--asl2struc", help="ASL->Structural transformation matrix", type="matrix", default=None)
        group.add_option("--wm-seg", help="White matter segmentation of structural image in structural space", type="image", default=None)
        group.add_option("--gm-seg", help="Grey matter segmentation of structural image in structural space", type="image", default=None)
        group.add_option("--csf-seg", help="CSF segmentation of structural image in structural space", type="image", default=None)
        group.add_option("--fslanat", help="FSL_ANAT output directory for structural information", default=None)
        group.add_option("--fastsrc", help="Images from a FAST segmentation - if not set FAST will be run on structural image")
        group.add_option("--struc2std", help="Structural to MNI152 linear registration (.mat)", type="matrix")
        group.add_option("--struc2std-warp", help="Structural to MNI152 non-linear registration (warp)", type="image")

        return [group, ]

def run(wsp):
    """
    Do initialization on supplied structural data - copy relevant image and do brain extraction

    FIXME copy across all supplied structural data
    """
    wsp.log.write("\nInitialising structural data\n")
    wsp.sub("structural")

    if wsp.fslanat:
        wsp.log.write(" - Using FSL_ANAT output directory for structural data: %s\n" % wsp.fslanat)
        biascorr = os.path.join(wsp.fslanat, "T1_biascorr")
        biascorr_brain = os.path.join(wsp.fslanat, "T1_biascorr_brain")
        if glob.glob(biascorr + ".*") and glob.glob(biascorr_brain + ".*"):
            wsp.log.write(" - Using bias-corrected structural images\n")
            wsp.structural.struc = Image(biascorr)
            wsp.structural.brain = Image(biascorr_brain)
        else:
            wsp.log.write(" - Using non bias-corrected structural images\n")
            wsp.structural.struc = Image(os.path.join(wsp.fslanat, "T1"))
            wsp.structural.brain = Image(os.path.join(wsp.fslanat, "T1_brain"))
    elif wsp.struc:
        wsp.log.write(" - Using structural image provided by user: %s\n" % wsp.struc.name)
        wsp.structural.struc = wsp.struc
        wsp.structural.brain = wsp.struc_brain
    #elif wsp.structural.struc_lores
    #    wsp.log.write("Low-resolution tructural image: %s\n" % wsp.structural.struc_lores.name)
    else:
        wsp.log.write(" - No structural data supplied - output will be ASL space only\n")

    if wsp.structural.struc is not None and wsp.structural.brain is None:
        wsp.log.write(" - Brain-extracting structural image\n")
        bet_result = fsl.bet(wsp.structural.struc, output=fsl.LOAD, seg=True, mask=True, log=wsp.fsllog)
        wsp.structural.brain = bet_result["output"]
        #wsp.structural.brain_mask = bet_result["output_mask"]

    if wsp.structural.brain is not None and wsp.structural.brain_mask is None:
        # FIXME - for now get the mask by binarising the brain image for compatibility with oxford_asl
        # although this gives slightly different results compared to using the mask returned by BET
        wsp.structural.brain_mask = Image((wsp.structural.brain.data != 0).astype(np.int32), header=wsp.structural.struc.header)

    if wsp.structural.struc is not None:
        segment(wsp)

def segment(wsp):
    """
    Segment the structural image
    """
    # Can override segmentation from user-specified maps
    wsp.structural.wm_seg = wsp.wm_seg
    wsp.structural.gm_seg = wsp.gm_seg
    wsp.structural.csf_seg = wsp.csf_seg

    if None in (wsp.structural.wm_seg, wsp.structural.gm_seg, wsp.structural.csf_seg):
        page = wsp.report.page("seg")
        page.heading("Segmentation of structural image")

        wsp.log.write("\nGetting structural segmentation\n")
        if wsp.fslanat:
            wsp.log.write(" - Using FSL_ANAT output from %s\n" % wsp.fslanat)
            page.text("Segmentation taken from FSL_ANAT output at ``%s``" % wsp.fslanat)
            wsp.structural.csf_pv = Image(os.path.join(wsp.fslanat, "T1_fast_pve_0"))
            wsp.structural.gm_pv = Image(os.path.join(wsp.fslanat, "T1_fast_pve_1"))
            wsp.structural.wm_pv = Image(os.path.join(wsp.fslanat, "T1_fast_pve_2"))

            try:
                wsp.structural.bias = Image(os.path.join(wsp.fslanat, "T1_fast_bias"))
                wsp.log.write(" - Bias field extracted sucessfully\n")
            except PathError:
                wsp.log.write(" - No bias field found")
        elif wsp.fastsrc:
            wsp.log.write(" - Using FAST output from %s\n" % wsp.fastsrc)
            page.text("Segmentation taken from FAST output at ``%s``" % wsp.fastsrc)
            wsp.structural.csf_pv = Image("%s_pve_0" % wsp.fastsrc)
            wsp.structural.gm_pv = Image("%s_pve_1" % wsp.fastsrc)
            wsp.structural.wm_pv = Image("%s_pve_2" % wsp.fastsrc)
        elif wsp.structural.struc:
            wsp.log.write(" - Running FAST\n")
            page.text("FAST run to segment structural image")
            fast_result = fsl.fast(wsp.structural.brain, out=fsl.LOAD, log=wsp.fsllog)
            wsp.structural.csf_pv = fast_result["out_pve_0"]
            wsp.structural.gm_pv = fast_result["out_pve_1"]
            wsp.structural.wm_pv = fast_result["out_pve_2"]
            #wsp.bias_struc = fast_result["fast_bias"]
        else:
            raise ValueError("No structural data provided - cannot segment")

        if wsp.structural.csf_seg is None:
            wsp.structural.csf_seg = Image((wsp.structural.csf_pv.data > 0.5).astype(np.int32), header=wsp.structural.struc.header)
        if wsp.structural.gm_seg is None:
            wsp.structural.gm_seg = Image((wsp.structural.gm_pv.data > 0.5).astype(np.int32), header=wsp.structural.struc.header)
        if wsp.structural.wm_seg is None:
            wsp.structural.wm_seg = Image((wsp.structural.wm_pv.data > 0.5).astype(np.int32), header=wsp.structural.struc.header)

        page.heading("Segmentation image", level=1)
        page.text("CSF partial volume")
        page.image("csf_pv", LightboxImage(wsp.structural.csf_pv, bgimage=wsp.structural.brain))
        page.text("Grey matter partial volume")
        page.image("gm_pv", LightboxImage(wsp.structural.gm_pv, bgimage=wsp.structural.brain))
        page.text("White matter partial volume")
        page.image("wm_pv", LightboxImage(wsp.structural.wm_pv, bgimage=wsp.structural.brain))
    else:
        # User-supplied segmentation - treat as PV as well
        wsp.structural.csf_pv = wsp.structural.csf_seg
        wsp.structural.gm_pv = wsp.structural.gm_seg
        wsp.structural.wm_pv = wsp.structural.wm_seg



"""
OXASL - BASIL: Quantification using Bayesian model fitting

The BASIL module is a little more complex than the other Workspace based
modules because of the number of options available and the need for flexibility
in how the modelling steps are run.

The main function is ``basil`` which performs model fitting on ASL data
in the Workspace ``asldata`` attribute.

    wsp = Workspace()
    wsp.asldata = AslImage("asldata.nii.gz", tis=[1.6,])
    wsp.infertiss = True
    basil.run(wsp.sub("basil"))
    wsp.basil.finalstep.mean_ftiss.save("mean_ftiss.nii.gz")

Because of the number of options possible for the modelling process, the
workspace attribute ``basil_options`` can be set as a dictionary of extra
options relevant only to Basil:

    wsp = Workspace()
    wsp.asldata = AslImage("asldata.nii.gz", tis=[1.6,])
    wsp.basil_options = {"infertiss" : True, "spatial" : True}
    basil.run(wsp.sub("basil"))
    wsp.basil.finalstep.mean_ftiss.save("mean_ftiss.nii.gz")

All options specified in basil_options are either consumed by Basil, or
if not passed directly to the model.

Copyright (c) 2008-2020 Univerisity of Oxford
"""
"""
OXASL - BASIL: Quantification using Bayesian model fitting

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from oxasl.options import OptionCategory, OptionGroup

from . import fabber_method, svb_method, vaby_method, multistep_fit

class Options(OptionCategory):
    """
    Options for corrections of the input data
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "corrections")

    def groups(self, parser):
        ret = []

        g = OptionGroup(parser, "BASIL quantification")
        g.add_option("--basil-mask", help="Masking policy to use for model fitting. Does not affect analysis mask used in rest of pipeline. 'dilate' means dilate the default analysis mask. 'none' means use no masking",
                     type="choice", choices=["default", "dilated", "none"])
        g.add_option("--basil-options", "--fit-options", help="File containing additional options for Basil model fitting", type="optfile", default=None)
        g.add_option("--basil-method", help="Model fitting method", type="choice", choices=["fabber", "vaby", "svb"])
        ret.append(g)

        g = OptionGroup(parser, "BASIL partial volume correction (PVEc)")
        g.add_option("--pvcorr", help="Apply PVEc using FAST estimates taken from --fslanat dir", action="store_true", default=False)
        g.add_option("--pvgm", help="GM PV estimates in ASL space (apply PVEc only, don't estimate PVs)", type="image", default=None)
        g.add_option("--pvwm", help="As above, WM PV estimates in ASL space", type="image", default=None)
        g.add_option("--pvcsf", help="As above, CSF PV estimates in ASL space", type="image", default=None)
        ret.append(g)

        return ret

def run(wsp):
    method = wsp.ifnone("basil_method", "fabber")
    wsp.log.write(" - Running Bayesian model fitting using method '%s'\n" % method)
    if method == "fabber":
        fabber_method.run(wsp)
    elif method == "svb":
        svb_method.run(wsp)
    elif method == "vaby":
        vaby_method.run(wsp)
    else:
        raise ValueError("Unknown BASIL method: %s" % method)



"""
Python implementation of EPI registration
"""
import os
import sys

import numpy as np

import fsl.wrappers as fsl

from oxasl import struc

def epi_reg(wsp, epi_img, use_fmap=False, **kwargs):
    """
    Do EPI registration
    """
    struc.segment(wsp)
    wmseg = wsp.structural.wm_seg
    bbr_sch = os.path.join(os.environ["FSLDIR"], "etc/flirtsch/bbr.sch")

    if wsp.asl2struc is None:
        # Do pre-alignment in the same was as epi_reg
        wsp.asl2struc = fsl.flirt(epi_img, ref=wsp.structural.brain, omat=fsl.LOAD, dof=6, log=wsp.fsllog)["omat"]

    if not use_fmap:
        wsp.log.write(" - Running BBR\n")
        trans = fsl.flirt(epi_img, ref=wsp.structural.struc, dof=6, cost="bbr", wmseg=wmseg, init=wsp.asl2struc, omat=fsl.LOAD, out=fsl.LOAD, schedule=bbr_sch, log=wsp.fsllog)["omat"]
        out_img = fsl.applywarp(epi_img, ref=wsp.structural.struc, out=fsl.LOAD, premat=trans, interp="spline", log=wsp.fsllog)["out"]
        return {"out.nii.gz" : out_img, "out" : trans}
    else:
        dirmap = {
            "x" : (1, "x"),
            "y" : (2, "y"),
            "z" : (3, "z"),
            "-x" : (-1, "x-"),
            "-y" : (-2, "y-"),
            "-z" : (-3, "z-"),
            "x-" : (-1, "x-"),
            "y-" : (-2, "y-"),
            "z-" : (-3, "z-"),
        }
        pedir, fdir = dirmap.get(wsp.pedir, (None, None))
        if pedir is None:
            raise ValueError("Invalid phase encode direction specified: %s" % wsp.pedir)

        if wsp.nofmapreg:
            wsp.fmap2struc = np.identity(4)
            wsp.fmapmag_struc = wsp.fmapmag
        else:
            # Register fmap to structural image
            wsp.log.write(" - Registering fieldmap to structural\n")
            wsp.fmap2struc = fsl.flirt(wsp.fmapmagbrain, ref=wsp.structural.brain, dof=6, omat=fsl.LOAD)["omat"]
            flirt_result = fsl.flirt(wsp.fmapmag, ref=wsp.structural.struc, dof=6, init=wsp.fmap2struc, omat=fsl.LOAD, out=fsl.LOAD, nosearch=True)
            wsp.fmap2struc = flirt_result["omat"]
            wsp.fmapmag_struc = flirt_result["out"]

        # Unmask the fieldmap (necessary to avoid edge effects)
        wsp.fmap_mask = fsl.fslmaths(wsp.fmapmagbrain).abs().bin().run()
        wsp.fmap_mask = fsl.fslmaths(wsp.fmap).abs().bin().mul(wsp.fmap_mask).run()

        # The direction here should take into account the initial affine (it needs to be the direction in the EPI)
        wsp.fmap_unmasked = fsl.fugue(loadfmap=wsp.fmap, mask=wsp.fmap_mask, unmaskfmap=True, savefmap=fsl.LOAD, unwarpdir=fdir)["out"]

        # The following is a NEW HACK to fix extrapolation when fieldmap is too small
        wsp.fmap_struc_pad0 = fsl.applywarp(wsp.fmap_unmasked, ref=wsp.structural.struc, premat=wsp.fmap2struc, out=fsl.LOAD)["out"]
        wsp.fmap_struc_innermask = fsl.fslmaths(wsp.fmap_struc_pad0).abs().bin().run()
        wsp.fmap_struc_dilated = fsl.fugue(loadfmap=wsp.fmap_struc_pad0, mask=wsp.fmap_struc_innermask, unmaskfmap=True, unwarpdir=fdir, savefmap=fsl.LOAD)["savefmap"]
        wsp.fmap_struc = wsp.fmap_struc_dilated

        # Run bbr with fieldmap
        wsp.log.write("Running BBR with fieldmap\n")
        if not wsp.epi_reg_use_weighting:
            refweight = None

        wsp.reg.epi2struc = fsl.flirt(epi_img, ref=wsp.structural.struc, dof=6, cost="bbr", wmseg=wmseg, init=wsp.reg.asl2struc, omat=fsl.LOAD, schedule=bbr_sch, echospacing=wsp.echospacing, pedir=pedir, fieldmap=wsp.fmap_struc, refweight=refweight)["omat"]

        # Make equivalent warp fields
        wsp.log.write("Making warp fields and applying registration to EPI series\n")
        wsp.reg.struc2epi = np.linalg.inv(wsp.reg.epi2struc)
        fsl.concatxfm(wsp.reg.struc2epi, wsp.fmap2struc, outmat=fsl.LOAD)
        fsl.applywarp(wsp.fmap, ref=epi_img, premat=wsp.fmap2epi, out=fsl.LOAD)
        wsp.fmap2epi_mask = fsl.fslmaths(wsp.fmap2epi).abs().bin().run()
        # ${vout}_fieldmaprads2epi -abs -bin ${vout}_fieldmaprads2epi_mask
        ret = fsl.fugue(loadfmap=wsp.fmap2epi, mask=wsp.fmap2epi_mask, saveshift=fsl.LOAD, unmaskshift=True, dwell=wsp.dwell, unwarpdir=fdir)
        print(ret)
        wsp.fmap2epi_shift = ret["fieldmaprads2epi_shift"]
        ret = fsl.convertwarp(wsp.structural.struc, s=wsp.fmap2epi_shift, postmat=wsp.epi2struc, out=fsl.LOAD, shiftdir=fdir, relout=True)
        print(ret)
        warp = ret["out"]
        ret = fsl.applywarp(epi_img, ref=wsp.structural.struc, out=fsl.LOAD, warp1=warp, interp="spline", rel=True)
        print(ret)
        return ret["out"], warp
        # if gdcwarp:
        #     ${FSLDIR}/bin/convertwarp --ref=${vepi} --warp1=${vout}_warp --midmat=${vout}_inv.mat --warp2=${gdcwarp}  --out=${vepi}2${vepi}_undistorted_warp --relout
        #     ${FSLDIR}/bin/applywarp -i ${vepi} -r ${vepi} -o ${vepi}_fully_undistorted -w ${vepi}2${vepi}_undistorted_warp --interp=spline --rel
        #     ${FSLDIR}/bin/flirt -ref ${vrefhead} -in ${vepi}_fully_undistorted -dof 6 -cost bbr -wmseg ${vout}_fast_wmseg -init ${vout}.mat -omat ${vout}.mat -out ${vout} -schedule ${FSLDIR}/etc/flirtsch/bbr.sch
        #     ${FSLDIR}/bin/convertwarp --ref=${vrefhead} --warp1=${vout}_warp --midmat=${vout}_inv.mat --warp2=${gdcwarp} --postmat=${vout}.mat --out=${vout}_warp --relout
        #     ${FSLDIR}/bin/applywarp -i ${vepi} -r ${vrefhead} -o ${vout} -w ${vout}_warp --interp=spline --rel

def main():
    """ Simple test case """
    from fsl.data.image import Image
    from oxasl import Workspace
    wsp = Workspace(savedir="epi_py_out", debug=True)
    wsp.sub("reg")
    wsp.aslref = Image(sys.argv[1])
    wsp.fslanat = sys.argv[2]
    struc.init(wsp)

    result = epi_reg(wsp, wsp.aslref)
    wsp.imgreg, wsp.img2struc = result["out.nii.gz"], result["out"]

if __name__ == "__main__":
    main()



"""
OXASL - Output module

Copyright (c) 2008-2020 Univerisity of Oxford
"""
import itertools

import numpy as np

from fsl.data.image import Image

from oxasl.options import OptionCategory, OptionGroup
from oxasl import reg, calibration
from oxasl.reporting import LightboxImage

class Options(OptionCategory):
    """
    OptionCategory which contains options for output files to keep
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "output")

    def groups(self, parser):
        groups = []

        g = OptionGroup(parser, "Output options")
        g.add_option("--save-input", help="Save unprocessed input data", action="store_true", default=False)
        g.add_option("--save-preproc", help="Save preprocessed input data", action="store_true", default=False)
        g.add_option("--save-corrected", help="Save corrected input data (distortion/moco/sensitivity corrected)", action="store_true", default=False)
        g.add_option("--save-corrections", help="Save intermediate correction data (e.g. fieldmaps, moco)", action="store_true", default=False)
        g.add_option("--save-reg", help="Save registration information (transforms etc)", action="store_true", default=False)
        g.add_option("--save-filter", help="Save data filtering output", action="store_true", default=False)
        g.add_option("--save-quantification", help="Save quantification output", action="store_true", default=False)
        g.add_option("--save-calib", help="Save calibration output", action="store_true", default=False)
        g.add_option("--save-all", help="Save all output (enabled when --debug specified)", action="store_true", default=False)
        g.add_option("--output-stddev", "--output-std", help="Output standard deviation of estimated variables", action="store_true", default=False)
        g.add_option("--output-var", "--vars", help="Output variance of estimated variables", action="store_true", default=False)
        g.add_option("--output-residuals", help="Output residuals (model fit - actual data)", action="store_true", default=False)
        g.add_option("--output-mni", help="Output in MNI standard space", action="store_true", default=False)
        g.add_option("--output-custom", help="Output in custom space (provide path to reference image in space)", type=str)
        g.add_option("--output-custom-mat", help="(Optional) FLIRT transformation from structural space to custom space. " +
                        "If not provided, will FLIRT registration from structural to --output-custom will be used.", type=str)
        g.add_option("--no-report", dest="save_report", help="Don't try to generate an HTML report", action="store_false", default=True)
        groups.append(g)

        return groups

OUTPUT_ITEMS = {
    "ftiss" : ("perfusion", 6000, True, "ml/100g/min", "30-50", "10-20"),
    "fblood" : ("aCBV", 100, True, "ml/100g/min", "", ""),
    "delttiss" : ("arrival", 1, False, "s", "", ""),
    "fwm" : ("perfusion_wm", 6000, True, "ml/100g/min", "", "10-20"),
    "deltwm" : ("arrival_wm", 1, False, "s", "", ""),
    "noise_means" : ("noise", 1, False, "", "", ""),
    "modelfit" : ("modelfit", 1, False, "", "", ""),
    "modelfit_mean" : ("modelfit_mean", 1, False, "", "", ""),
    "residuals" : ("residuals", 1, False, "", "", ""),
    "asldata_diff" : ("asldata_diff", 1, False, "", "", ""),
    "T_exch" : ("texch", 1, False, "", "", ""),
}

def run(wsp):
    wsp.log.write("\nGenerating output images\n")

    for quantify_wsp in wsp.quantify_wsps:
        try:
            quantify_name = quantify_wsp[quantify_wsp.index("_")+1:]
            output_wsp = "output_%s" % quantify_name
        except ValueError:
            quantify_name = ""
            output_wsp = "output"
        _output_from_quantification(quantify_name, getattr(wsp, quantify_wsp), wsp.sub(output_wsp))

def _output_from_quantification(quantify_name, quantify_wsp, output_wsp):
    """
    """
    output_wsp.full_region_analysis = quantify_wsp.full_region_analysis
    if quantify_wsp.image_space is None:
        _output_native(output_wsp.sub("native"), quantify_name, quantify_wsp)
        _output_trans(output_wsp)
    else:
        _output_native(output_wsp, quantify_name, quantify_wsp)

def _output_native(wsp, quantify_name, quantify_wsp):
    """
    Create output images from quantification

    This includes basic sanity-processing (removing negatives and NaNs) plus
    calibration using existing M0

    :param wsp: Workspace object for output
    :param quantify_wsp: Workspace in which quantification has been run. The ``finalstep``
                      attribute is expected to point to the final output workspace
    """
    wsp.log.write(" - Generating native (ASL) space output\n")
    # Output the differenced data averaged across repeats for kinetic curve comparison
    # with the model
    if wsp.asldata.iaf in ("tc", "ct", "diff"):
        wsp.diffdata_mean = wsp.asldata.diff().mean_across_repeats()

    # Save the analysis mask
    wsp.mask = quantify_wsp.analysis_mask

    # Output model fitting results
    prefixes = ["", "mean"]
    if wsp.output_stddev:
        prefixes.append("std")
    if wsp.output_var or wsp.region_analysis:
        prefixes.append("var")
    for fabber_name, oxasl_output in OUTPUT_ITEMS.items():
        for prefix in prefixes:
            is_variance = prefix == "var"
            if is_variance:
                # Variance is not output by Fabber natively so we get it by
                # squaring the standard deviation. We also pass the flag
                # to the calibration routine so it can square the correction
                # factors
                fabber_output = "std_%s" % fabber_name
            elif prefix:
                fabber_output = "%s_%s" % (prefix, fabber_name)
            else:
                fabber_output = fabber_name

            img = quantify_wsp.finalstep.ifnone(fabber_output, None)
            if img is not None:
                # Make negative/nan values = 0 and ensure masked value zeroed
                # since quantification may use a different fitting mask to the pipeline mask
                data = np.copy(img.data)
                data[~np.isfinite(data)] = 0
                data[img.data < 0] = 0
                data[wsp.mask.data == 0] = 0
                img = Image(data, header=img.header)
                name, multiplier, calibrate, units, normal_gm, normal_wm = oxasl_output
                if prefix and prefix != "mean":
                    name = "%s_%s" % (name, prefix)

                # Variance is not output by Fabber natively so we get it by
                # squaring the standard deviation.
                if is_variance:
                    img = Image(np.square(img.data), header=img.header)
                setattr(wsp, name, img)

                if calibrate:
                    for method in wsp.calibration.calib_method:
                        if method != "prequantified":
                            calib_wsp = getattr(wsp.calibration, method)
                            img_calib = calibration.run(calib_wsp, img, multiplier=multiplier, var=is_variance)
                            sub_wsp_name = "calib_%s" % method
                            calib_output_wsp = getattr(wsp, sub_wsp_name)
                            if calib_output_wsp is None:
                                calib_output_wsp = wsp.sub(sub_wsp_name)
                            setattr(calib_output_wsp, name, img_calib)
                            output_report(calib_output_wsp, name, units, normal_gm, normal_wm, method, quantify_name)
                        else:
                            output_report(wsp, name, units, normal_gm, normal_wm, method, quantify_name)
                else:
                    output_report(wsp, name, units, normal_gm, normal_wm, "none", quantify_name)

def output_report(wsp, name, units, normal_gm, normal_wm, calib_method="none", quantify_name=""):
    """
    Create report pages from output data

    :param wsp: Workspace object containing output
    """
    report = wsp.report
    if not quantify_name:
        quantify_name = "standard"

    img = getattr(wsp, name)
    if img is not None and img.ndim == 3:
        page = report.page("%s_%s_%s" % (name, calib_method, quantify_name))
        page.heading("Output image: %s (calibration: %s, quantification: %s)" % (name, calib_method, quantify_name))
        if calib_method != "none":
            alpha = wsp.ifnone("calib_alpha", 1.0 if wsp.asldata.iaf in ("ve", "vediff") else 0.85 if wsp.asldata.casl else 0.98)
            page.heading("Calibration", level=1)
            page.text("Calibration method: %s" % calib_method)
            page.text("Inversion efficiency: %f" % alpha)

        page.heading("Metrics", level=1)
        data = img.data
        roi = reg.change_space(wsp, wsp.mask, img).data
        table = []
        table.append(["Mean within mask", "%.4g %s" % (np.mean(data[roi > 0.5]), units), ""])
        if wsp.structural.struc is not None:
            table.append(["GM mean", "%.4g %s" % (np.mean(data[wsp.rois.gm_asl.data > 0]), units), normal_gm])
            table.append(["Pure GM mean", "%.4g %s" % (np.mean(data[wsp.rois.pure_gm_asl.data > 0]), units), normal_gm])
            table.append(["Cortical GM mean", "%.4g %s" % (np.mean(data[wsp.rois.cortical_gm_asl.data > 0]), units), normal_gm])
            table.append(["WM mean", "%.4g %s" % (np.mean(data[wsp.rois.wm_asl.data > 0]), units), normal_wm])
            table.append(["Pure WM mean", "%.4g %s" % (np.mean(data[wsp.rois.pure_wm_asl.data > 0]), units), normal_wm])
            table.append(["Cerebral WM mean", "%.4g %s" % (np.mean(data[wsp.rois.cerebral_wm_asl.data > 0]), units), normal_wm])

        page.table(table, headers=["Metric", "Value", "Typical"])

        page.heading("Image", level=1)
        page.image("%s_%s_%s_img" % (name, quantify_name, calib_method), LightboxImage(img, zeromask=False, mask=wsp.mask, colorbar=True))

def __output_trans_helper(wsp):
    """
    Generator to provide all the combinations of variables for the output_
    trans() function. Note that 4D data and upwards will be skipped. 
    """

    # We loop over these combinations of variables for each output space
    # (structural, standard, custom)
    suffixes = ("", "_std", "_var")
    outputs = (
        "perfusion", "aCBV", "arrival", "perfusion_wm", 
        "arrival_wm", "modelfit", "modelfit_mean", 
        "residuals", "texch", "mask"
    )

    for suff, out in itertools.product(suffixes, outputs):
        data = getattr(wsp.native, out + suff)
        # Don't transform 4D output (e.g. modelfit) - too large!
        if (data is not None) and data.ndim == 3:
            yield suff, out, data

def _output_trans(wsp):
    """
    Create transformed output, i.e. in structural and/or standard space
    """
    output_spaces = []
    if wsp.output_struc and wsp.reg.asl2struc is not None:
        output_spaces.append(("struc", "structural"))
    
    if wsp.output_mni:
        if wsp.reg.struc2asl is None:
            wsp.log.write(" - WARNING: No structural registration - cannot output in standard space\n")
        else:
            output_spaces.append(("std", "standard (MNI)"))
    
    if wsp.output_custom:
        output_spaces.append(("custom", "user-defined custom"))

    for space, name in output_spaces:
        wsp.log.write(" - Generating output in %s space\n" % name)
        output_wsp = wsp.sub(space)
        for suffix, output, native_data in __output_trans_helper(wsp): 
            setattr(output_wsp, output + suffix, reg.change_space(wsp, native_data, space, mask=(output == 'mask')))
            for method in wsp.calibration.calib_method:
                if method != "prequantified":
                    sub_wsp_name = "calib_%s" % method
                    native_calib_output_wsp = getattr(wsp.native, sub_wsp_name)
                    native_calib_data = getattr(native_calib_output_wsp, output + suffix)
                    calib_output_wsp = output_wsp.sub(sub_wsp_name)
                    setattr(calib_output_wsp, output + suffix, reg.change_space(wsp, native_calib_data, space, mask=(output == 'mask')))
                else:
                    native_data = getattr(wsp.native, output + suffix)
                    setattr(output_wsp, output + suffix, reg.change_space(wsp, native_data, space, mask=(output == 'mask')))



#!/bin/env python
"""
OXASL - Quantification module

Fits a kinetic model to differenced ASL data

Copyright (c) 2008-2020 Univerisity of Oxford
"""

from oxasl import basil

try:
    import oxasl_multite
except ImportError:
    oxasl_multite = None

try:
    import oxasl_ve
except ImportError:
    oxasl_ve = None

def run(wsp):
    wsp.quantify_wsps = []
    quantify = _get_quantify_method(wsp)
    wsp.log.write("\nQuantifying perfusion from ASL data\n")
    quantify(wsp)

def _already_quantified(wsp):
    # Make fake basil-like directory for already quantified data
    wsp.log.write(" - Skipping quantification, data is already quantified\n")
    quantify_wsp = wsp.sub("prequant")
    # Two possible locations for compatibility
    if wsp.rois is not None and wsp.rois.mask is not None:
        quantify_wsp.analysis_mask = wsp.rois.mask
    else:
        quantify_wsp.analysis_mask = wsp.mask
    finaldir = quantify_wsp.sub("finalstep")
    finaldir.mean_ftiss = wsp.asldata
    wsp.quantify_wsps.append("prequant")
    return quantify_wsp

def _get_quantify_method(wsp):
    if wsp.asldata.iaf == "quant":
        return _already_quantified
    elif wsp.asldata.iaf in ("tc", "ct", "diff", "hadamard"):
        if wsp.asldata.ntes == 1:
            return basil.run
        elif oxasl_multite is None:
            raise ValueError("Multi-TE data supplied but oxasl_multite is not installed")
        else:
            return oxasl_multite.run
    elif wsp.asldata.iaf == "ve":
        if oxasl_ve is None:
            raise ValueError("VE data supplied but oxasl_ve is not installed")
        else:
            return oxasl_ve.run
    else:
        raise ValueError("No known quantification method for data of type %s" % wsp.asldata.iaf)



"""This module provides wrapper functions for the FSL `FNIRTFILEUTILS
<https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FNIRT/>`_ tool

.. autosummary::
   :nosignatures:

   fnirtfileutils
"""

import fsl.utils.assertions as asrt
from fsl.wrappers import wrapperutils  as wutils

@wutils.fileOrImage('src', 'ref', 'out', 'jac')
@wutils.fslwrapper
def fnirtfileutils(src, **kwargs):
    """Wrapper for the ``fnirt`` command.
    
    Compulsory arguments (You MUST set one or more of):
	    -i,--in		filename of input coefficient volume (to be converted)

    Optional arguments (You may optionally specify one or more of):
        -r,--ref	filename for reference volume
        -o,--out	filename for output (field/coef) volume - uses relative warp convention
        -f,--outformat	Output format [field spline], default=field
        -w,--warpres	Warp resolution (mm), only relevant when --outformat=spline
        -k,--knotspace	Knot-spacing (voxels), only relevant when --outformat=spline
        -j,--jac	filename for output (jacobian map) volume
        -a,--withaff	If set, the affine transform is included in the field/jacobian
        -v,--verbose	switch on diagnostic messages
        -h,--help	display this message
    """

    asrt.assertIsNifti(src)

    cmd = ['fnirtfileutils', '--in={}'.format(src)]
    cmd += wutils.applyArgStyle('--=', **kwargs)

    return cmd




"""
oxasl.gui.structure_tab.py

Copyright (c) 2019 University of Oxford
"""
import os

import wx

from oxasl.gui.widgets import TabPage, OptionError

class StructureTab(TabPage):
    """
    Tab page containing options for providing structural data
    """

    EXISTING_FSLANAT = 0
    NEW_FSLANAT = 1
    INDEP_STRUC = 2
    NONE = 3

    TRANS_MATRIX = 0
    TRANS_IMAGE = 1
    TRANS_FSLANAT = 2

    def __init__(self, parent, idx, n):
        TabPage.__init__(self, parent, "Structure", idx, n, name="structure")

        self.section("Structure")

        self.struc_ch = wx.Choice(self, choices=[
            "Existing FSL_ANAT output",
            "Run FSL_ANAT on structural image",
            "Independent structural data",
            "None"
        ])
        self.struc_ch.SetSelection(self.NONE)
        self.struc_ch.Bind(wx.EVT_CHOICE, self._changed)
        self.struc_ch.span = 2
        self.pack("Structural data from", self.struc_ch)

        self.fsl_anat_picker = self.file_picker("Existing FSL_ANAT directory", pick_dir=True)
        self.struc_image_picker = self.file_picker("Structural Image")
        self.brain_image_picker = self.file_picker("Brain image", optional=True)

        self.section("Registration")

        self.transform_choices = ["Use matrix", "Use warp image", "Use FSL_ANAT"]

        self.transform_cb = wx.CheckBox(self, label="Transform to standard space")
        self.transform_cb.Bind(wx.EVT_CHECKBOX, self._changed)
        self.pack("", self.transform_cb, enable=False)

        self.sizer.AddGrowableCol(2, 1)
        self.SetSizer(self.sizer)
        self.next_prev()

    def options(self):
        options = {}
        if self.struc_ch.GetSelection() == self.INDEP_STRUC:
            options["struc"] = self.image("Structural data", self.struc_image_picker.GetPath())
            if self.brain_image_picker.checkbox.IsChecked():
                options["struc_brain"] = self.image("Structural brain", self.brain_image_picker.GetPath())
        elif self.struc_ch.GetSelection() == self.EXISTING_FSLANAT:
            options["fslanat"] = self.fsl_anat_picker.GetPath()
            if not os.path.exists(options["fslanat"]):
                raise OptionError("FSL_ANAT directory does not exist")
        # FIXME new FSL_ANAT

        if self.transform_cb.IsChecked():
            options["output_std"] = True
        return options

    def update(self):
        mode = self.struc_ch.GetSelection()
        self.fsl_anat_picker.Enable(mode == self.EXISTING_FSLANAT)
        self.struc_image_picker.Enable(mode in (self.NEW_FSLANAT, self.INDEP_STRUC))

        self.brain_image_picker.checkbox.Enable(mode == self.INDEP_STRUC)
        self.brain_image_picker.Enable(mode == self.INDEP_STRUC and self.brain_image_picker.checkbox.IsChecked())

        TabPage.update(self)
        